# -*- coding: utf-8 -*-
# Copyright 2024 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function


__metaclass__ = type

"""
The Route_maps parser templates file. This contains
a list of parser definitions and associated functions that
facilitates both facts gathering and native command generation for
the given network resource.
"""

import re

from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.rm_base.network_template import (
    NetworkTemplate,
)


class Route_mapsTemplate(NetworkTemplate):
    def __init__(self, lines=None, module=None):
        super(Route_mapsTemplate, self).__init__(lines=lines, tmplt=self, module=module)

    # fmt: off
    PARSERS = [
        {
            "name": "condition",
            "getval": re.compile(
                r"""^dummy-regex
                $""", re.VERBOSE,
            ),
            "setval": "{{ condition_type }}{{ ' ' + condition if condition_type!='else' }}{{ ' then' if condition_type!='else' }}",
            "result": {},
        },
        {
            "name": "add.eigrp_metric",
            "getval": re.compile(
                r"""
                \s*add\seigrp-metric
                (\s(?P<bandwidth>\d+))?
                (\s(?P<delay>\d+))?
                (\s(?P<reliability>\d+))?
                (\s(?P<effective_bandwith>\d+))?
                (\s(?P<max_transmission>\d+))?
                $""", re.VERBOSE,
            ),
            "setval": "add eigrp-metric {{ add.eigrp_metric.bandwidth|string }}"
            " {{ add.eigrp_metric.delay|string }} {{ add.eigrp_metric.reliability|string }}"
            " {{ add.eigrp_metric.effective_bandwith|string }} {{ add.eigrp_metric.max_transmission|string }}",
            "result": {
                "policies": {
                    "add": {
                        "eigrp_metric": {
                            "bandwidth": "{{ bandwidth }}",
                            "delay": "{{ delay }}",
                            "reliability": "{{ reliability }}",
                            "effective_bandwith": "{{ effective_bandwith }}",
                            "max_transmission": "{{ max_transmission }}",
                        },
                    },
                },
            },
        },
        {
            "name": "add.rip_metric",
            "getval": re.compile(
                r"""
                \s*add\srip_metric
                (\s(?P<rip_metric>\d+))?
                $""", re.VERBOSE,
            ),
            "setval": "add rip-metric {{ add.rip_metric|string }}",
            "result": {
                "policies": {
                    "add": {
                        "rip_metric": "{{ bandwidth }}",
                    },
                },
            },
        },
        {
            "name": "apply",
            "getval": re.compile(
                r"""
                \s*apply
                (\s(?P<route_policy>\S+))
                (\s(?P<route_policy_input>.+))?
                $""", re.VERBOSE,
            ),
            "setval": "apply"
            "{{ (' ' + apply.route_policy) if apply.route_policy is defined else '' }}"
            "{{ (' ' + apply.route_policy_input) if apply.route_policy_input is defined else '' }}",
            "result": {
                "policies": {
                    "apply": [
                        {
                            "route_policy": "{{ route_policy }}",
                            "route_policy_input": "{{ route_policy_input }}",
                        },
                    ],
                },
            },
        },
        {
            "name": "drop",
            "getval": re.compile(
                r"""
                \s*drop
                $""", re.VERBOSE,
            ),
            "setval": "drop",
            "result": {
                "policies": {
                    "drop": True,
                },
            },
        },
        {
            "name": "pass",
            "getval": re.compile(
                r"""
                \s*pass
                $""", re.VERBOSE,
            ),
            "setval": "pass",
            "result": {
                "policies": {
                    "pass": True,
                },
            },
        },
        {
            "name": "prepend",
            "getval": re.compile(
                r"""
                \s*prepend
                (\sas-path\s(?P<as_path>\d+))?
                (\s(?P<most_recent>most-recent))?
                (\s(?P<own_as>own-as))?
                (\s(?P<number_of_times>\d+))?
                $""", re.VERBOSE,
            ),
            "setval": "prepend"
            "{{ (' as-path ' + prepend.as_path|string) if prepend.as_path is defined else '' }}"
            "{{ (' most-recent') if prepend.most_recent is defined else '' }}"
            "{{ (' own-as') if prepend.own_as|d(False) else '' }}"
            "{{ (' ' + prepend.number_of_times|string) if prepend.number_of_times is defined else '' }}",
            "result": {
                "policies": {
                    "prepend": {
                        "as_path": "{{ as_path }}",
                        "most_recent": "{{ not not most_recent }}",
                        "own_as": "{{ not not own_as }}",
                        "number_of_times": "{{ number_of_times }}",
                    },
                },
            },
        },
        {
            "name": "suppress_route",
            "getval": re.compile(
                r"""
                \s*suppress-route
                $""", re.VERBOSE,
            ),
            "setval": "suppress-route",
            "result": {
                "policies": {
                    "suppress_route": True,
                },
            },
        },
        {
            "name": "unsuppress_route",
            "getval": re.compile(
                r"""
                \s*unsuppress-route
                $""", re.VERBOSE,
            ),
            "setval": "unsuppress-route",
            "result": {
                "policies": {
                    "unsuppress_route": True,
                },
            },
        },
        {
            "name": "remove",
            "getval": re.compile(
                r"""
                \s*remove\sas-path
                (\s(?P<set>private-as))
                (\s(?P<entire_aspath>entire-aspath))?
                $""", re.VERBOSE,
            ),
            "setval": "remove as-path"
            "{{ (' private-as' ) if remove.set|d(False) else '' }}"
            "{{ (' entire-aspath' ) if remove.entire_aspath|d(False) else '' }}",
            "result": {
                "policies": {
                    "remove": {
                        "set": True,
                        "entire_aspath": "{{ not not entire_aspath }}",
                    },
                },
            },
        },
        {
            "name": "set.administrative_distance",
            "getval": re.compile(
                r"""
                \s*set\sadministrative-distance
                (\s(?P<administrative_distance>\d+))?
                $""", re.VERBOSE,
            ),
            "setval": "set administrative-distance {{ set.administrative_distance|string }}",
            "result": {
                "policies": {
                    "set": {
                        "administrative_distance": "{{ administrative_distance }}",
                    },
                },
            },
        },
        {
            "name": "set.local_preference",
            "getval": re.compile(
                r"""
                \s*set\slocal-preference\s(?P<increment>\+)?
                        (?P<decrement>\-)?
                        (?P<multiply>\*)?
                        (?P<metric_number>\d+)
                $""", re.VERBOSE,
            ),
            "setval": "{% for pref in set.local_preference %}"
            "set local-preference"
            "{{ ' *' if pref.multiply|d(False) else '' }}"
            "{{ ' +' if pref.increment|d(False) else '' }}"
            "{{ ' -' if pref.decrement|d(False) else '' }}"
            "{{ ' ' ~ pref.metric_number|string }}\n"
            "{% endfor %}",
            "result": {
                "policies": {
                    "set": {
                        "local_preference": [
                            {
                                "increment": "{{ not not increment }}",
                                "metric_number": "{{ metric_number}}",
                                "decrement": "{{ not not decrement }}",
                                "multiply": "{{ not not multiply }}",
                            },
                        ],
                    },
                },
            },
        },
        {
            "name": "set.aigp_metric",
            "getval": re.compile(
                r"""
                \s*set\saigp-metric
                (\s(?P<icrement>\+))?
                (\s(?P<decrement>\-))?
                (\s(?P<metric_number>\d+))?
                (\s(?P<igp_cost>igp-cost))?
                $""", re.VERBOSE,
            ),
            "setval": "set aigp-metric"
            "{{ (' +' ) if set.aigp_metric.icrement is defined else '' }}"
            "{{ (' -' ) if set.aigp_metric.decrement is defined else '' }}"
            "{{ (' ' +  set.aigp_metric.metric_number|string) if set.aigp_metric.metric_number is defined else '' }}"
            "{{ (' igp-cost' ) if set.aigp_metric.igp_cost is defined else '' }}",
            "result": {
                "policies": {
                    "set": {
                        "aigp_metric": {
                            "icrement": "{{ not not icrement }}",
                            "decrement": "{{ not not decrement }}",
                            "metric_number": "{{ metric_number }}",
                            "igp_cost": "{{ not not igp_cost }}",
                        },
                    },
                },
            },
        },
        {
            "name": "set.attribute_set",
            "getval": re.compile(
                r"""
                \s*set\sattribute-set\sname-string
                (\s(?P<attribute_set>\d+))?
                $""", re.VERBOSE,
            ),
            "setval": "set attribute-set name-string {{ set.attribute_set|string }}",
            "result": {
                "policies": {
                    "set": {
                        "attribute_set": "{{ attribute_set }}",
                    },
                },
            },
        },
        {
            "name": "set.c_multicast_routing",
            "getval": re.compile(
                r"""
                \s*set\sc-multicast-routing
                (\s(?P<bgp>bgp))?
                (\s(?P<pim>pim))?
                $""", re.VERBOSE,
            ),
            "setval": "set c-multicast-routing"
            "{{ (' bgp' ) if set.c_multicast_routing.bgp is defined else '' }}"
            "{{ (' pim' ) if set.c_multicast_routing.pim is defined else '' }}",
            "result": {
                "policies": {
                    "set": {
                        "c_multicast_routing": {
                            "bgp": "{{ not not bgp }}",
                            "pim": "{{ not not pim }}",
                        },
                    },
                },
            },
        },
        {
            "name": "set.community",
            "getval": re.compile(
                r"""
                \s*set\scommunity
                (\s(?P<community_name>(.*?)))?
                (\s(?P<additive>additive))?
                $""", re.VERBOSE,
            ),
            "setval": "set community"
            "{{ (' ' + set.community.community_name ) if set.community.community_name is defined else '' }}"
            "{{ (' additive') if set.community.additive|d(False) else '' }}",
            "result": {
                "policies": {
                    "set": {
                        "community": {
                            "community_name": "{{ community_name }}",
                            "additive": "{{ not not additive }}",
                        },
                    },
                },
            },
        },
        {
            "name": "set.core_tree",
            "getval": re.compile(
                r"""
                \s*set\score-tree
                (\s(?P<ingress_replication>ingress-replication))?
                (\s(?P<ingress_replication_default>ingress-replication-default))?
                (\s(?P<ingress_replication_partitioned>ingress-replication-partitioned))?
                (\s(?P<mldp>mldp))?
                (\s(?P<mldp_default>mldp-default))?
                (\s(?P<mldp_inband>mldp-inband))?
                (\s(?P<mldp_partitioned_mp2mp>mldp-partitioned-mp2mp))?
                (\s(?P<mldp_partitioned_p2mp>mldp-partitioned-p2mp))?
                (\s(?P<p2mp_te>p2mp-te))?
                (\s(?P<p2mp_te_default>p2mp-te-default))?
                (\s(?P<p2mp_te_partitioned>p2mp-te-partitioned))?
                (\s(?P<pim_default>pim-default))?
                (\s(?P<sr_p2mp>sr-p2mp))?
                $""", re.VERBOSE,
            ),
            "setval": "set core-tree"
            "{{ (' ingress-replication' ) if set.core_tree.ingress_replication|d(False) is defined else '' }}"
            "{{ (' ingress-replication-default' ) if set.core_tree.ingress_replication_default|d(False) is defined else '' }}"
            "{{ (' ingress-replication-partitioned' ) if set.core_tree.ingress_replication_partitioned|d(False) is defined else '' }}"
            "{{ (' mldp' ) if set.core_tree.mldp|d(False) is defined else '' }}"
            "{{ (' mldp-default' ) if set.core_tree.mldp_default|d(False) is defined else '' }}"
            "{{ (' mldp-inband' ) if set.core_tree.mldp_inband|d(False) is defined else '' }}"
            "{{ (' mldp-partitioned-mp2mp' ) if set.core_tree.mldp_partitioned_mp2mp|d(False) is defined else '' }}"
            "{{ (' mldp-partitioned-p2mp' ) if set.core_tree.mldp_partitioned_p2mp|d(False) is defined else '' }}"
            "{{ (' p2mp-te' ) if set.core_tree.p2mp_te|d(False) is defined else '' }}"
            "{{ (' p2mp-te-default' ) if set.core_tree.p2mp_te_default|d(False) is defined else '' }}"
            "{{ (' p2mp-te-partitioned' ) if set.core_tree.p2mp_te_partitioned|d(False) is defined else '' }}"
            "{{ (' pim-default' ) if set.core_tree.pim_default|d(False) is defined else '' }}"
            "{{ (' sr-p2mp' ) if set.core_tree.sr_p2mp|d(False) is defined else '' }}",
            "result": {
                "policies": {
                    "set": {
                        "core_tree": {
                            "ingress_replication": "{{ not not ingress_replication }}",
                            "ingress_replication_default": "{{ not not ingress_replication_default }}",
                            "ingress_replication_partitioned": "{{ not not ingress_replication_partitioned }}",
                            "mldp": "{{ not not mldp }}",
                            "mldp_default": "{{ not not mldp_default }}",
                            "mldp_inband": "{{ not not mldp_inband }}",
                            "mldp_partitioned_mp2mp": "{{ not not mldp_partitioned_mp2mp }}",
                            "mldp_partitioned_p2mp": "{{ not not mldp_partitioned_p2mp }}",
                            "p2mp_te": "{{ not not p2mp_te }}",
                            "p2mp_te_default": "{{ not not p2mp_te_default }}",
                            "p2mp_te_partitioned": "{{ not not p2mp_te_partitioned }}",
                            "pim_default": "{{ not not pim_default }}",
                            "sr_p2mp": "{{ not not sr_p2mp }}",
                        },
                    },
                },
            },
        },
        {
            "name": "set.dampening",
            "getval": re.compile(
                r"""
                \s*set\sdampening
                (\s(?P<halflife>\d+))?
                (\s(?P<suppress>\d+))?
                (\s(?P<reuse>\d+))?
                (\s(?P<max_suppress>\d+))?
                $""", re.VERBOSE,
            ),
            "setval": "set dampening"
            "{{ (' ' + set.dampening.halflife|string) if set.dampening.halflife is defined else '' }}"
            "{{ (' ' + set.dampening.max_suppress|string) if set.dampening.max_suppress is defined else '' }}"
            "{{ (' ' + set.dampening.reuse|string) if set.dampening.reuse is defined else '' }}"
            "{{ (' ' + set.dampening.suppress|string) if set.dampening.suppress is defined else '' }}",
            "result": {
                "policies": {
                    "set": {
                        "dampening": {
                            "halflife": "{{ halflife }}",
                            "max_suppress": "{{ max_suppress }}",
                            "reuse": "{{ reuse }}",
                            "suppress": "{{ suppress }}",
                        },
                    },
                },
            },
        },
        {
            "name": "set.downstream_core_tree",
            "getval": re.compile(
                r"""
                \s*set\sdownstream-core-tree
                (\s(?P<ingress_replication>ingress-replication))?
                (\s(?P<mldp>mldp))?
                (\s(?P<p2mp_te>p2mp-te))?
                (\s(?P<sr_p2mp>sr-p2mp))?
                $""", re.VERBOSE,
            ),
            "setval": "set downstream-core-tree"
            "{{ (' ingress-replication' ) if set.downstream_core_tree.ingress_replication|d(False) is defined else '' }}"
            "{{ (' mldp' ) if set.downstream_core_tree.mldp|d(False) is defined else '' }}"
            "{{ (' p2mp-te' ) if set.downstream_core_tree.p2mp_te|d(False) is defined else '' }}"
            "{{ (' sr-p2mp' ) if set.downstream_core_tree.sr_p2mp|d(False) is defined else '' }}",
            "result": {
                "policies": {
                    "set": {
                        "downstream_core_tree": {
                            "ingress_replication": "{{ not not ingress_replication }}",
                            "mldp": "{{ not not mldp }}",
                            "p2mp_te": "{{ not not p2mp_te }}",
                            "sr_p2mp": "{{ not not sr_p2mp }}",
                        },
                    },
                },
            },
        },
        {
            "name": "set.eigrp_metric",
            "getval": re.compile(
                r"""
                \s*set\seigrp-metric
                (\s(?P<bandwidth>\d+))?
                (\s(?P<delay>\d+))?
                (\s(?P<reliability>\d+))?
                (\s(?P<effective_bandwith>\d+))?
                $""", re.VERBOSE,
            ),
            "setval": "set eigrp-metric"
            "{{ (' ' + set.eigrp_metric.bandwidth|string) if set.eigrp_metric.bandwidth is defined else '' }}"
            "{{ (' ' + set.eigrp_metric.delay|string) if set.eigrp_metric.delay is defined else '' }}"
            "{{ (' ' + set.eigrp_metric.reliability|string) if set.eigrp_metric.reliability is defined else '' }}"
            "{{ (' ' + set.eigrp_metric.effective_bandwith|string) if set.eigrp_metric.effective_bandwith is defined else '' }}",
            "result": {
                "policies": {
                    "set": {
                        "eigrp_metric": {
                            "bandwidth": "{{ bandwidth }}",
                            "delay": "{{ delay }}",
                            "reliability": "{{ reliability }}",
                            "effective_bandwith": "{{ effective_bandwith }}",
                            "max_transmission": "{{ max_transmission }}",
                        },
                    },
                },
            },
        },
        {
            "name": "set.extcommunity",
            "getval": re.compile(
                r"""
                \s*set\sextcommunity
                (\ssoo\s(?P<soo>\S+))?
                (\srt\s(?P<rt>\S+))?
                (\sbandwidth\s(?P<bandwidth>\S+))?
                (\scolor\s(?P<color>\S+))?
                (\scost\s(?P<cost>\S+))?
                (\sredirect-to-rt\s(?P<redirect_to_rt>\([^\)]+\)))?
                (\sseg-nh\s(?P<seg_nh>\S+))?
                (\s(?P<additive>additive))?
                $""", re.VERBOSE,
            ),
            "setval": "set extcommunity"
            "{{ (' soo ' + set.extcommunity.soo ) if set.extcommunity.soo is defined else '' }}"
            "{{ (' rt ' + set.extcommunity.rt ) if set.extcommunity.rt is defined else '' }}"
            "{{ (' bandwidth ' + set.extcommunity.bandwidth ) if set.extcommunity.bandwidth is defined else '' }}"
            "{{ (' color ' + set.extcommunity.color ) if set.extcommunity.color is defined else '' }}"
            "{{ (' cost ' + set.extcommunity.cost ) if set.extcommunity.cost is defined else '' }}"
            "{{ (' redirect-to-rt ' + set.extcommunity.redirect_to_rt ) if set.extcommunity.redirect_to_rt is defined else '' }}"
            "{{ (' seg-nh ' + set.extcommunity.seg_nh ) if set.extcommunity.seg_nh is defined else '' }}"
            "{{ (' additive') if set.extcommunity.additive|d(False) else '' }}",
            "result": {
                "policies": {
                    "set": {
                        "extcommunity": {
                            "soo": "{{ soo }}",
                            "rt": "{{ rt }}",
                            "bandwidth": "{{ bandwidth }}",
                            "color": "{{ color }}",
                            "cost": "{{ cost }}",
                            "redirect_to_rt": "{{ redirect_to_rt }}",
                            "seg_nh": "{{ seg_nh }}",
                            "additive": "{{ not not additive }}",
                        },
                    },
                },
            },
        },
        {
            "name": "set.fallback_vrf_lookup",
            "getval": re.compile(
                r"""
                \s*set\sfallback-vrf-lookup
                $""", re.VERBOSE,
            ),
            "setval": "set fallback-vrf-lookup",
            "result": {
                "policies": {
                    "set": {
                        "fallback_vrf_lookup": True,
                    },
                },
            },
        },
        {
            "name": "set.flow_tag",
            "getval": re.compile(
                r"""
                \s*set\sflow-tag
                (\s(?P<flow_tag>\d+))
                $""", re.VERBOSE,
            ),
            "setval": "set flow-tag"
            "{{ (' ' + set.flow_tag|string) if set.flow_tag is defined else '' }}",
            "result": {
                "policies": {
                    "set": {
                        "flow_tag": "{{ flow_tag }}",
                    },
                },
            },
        },
        {
            "name": "set.forward_class",
            "getval": re.compile(
                r"""
                \s*set\sforward-class
                (\s(?P<flow_tag>\d+))
                $""", re.VERBOSE,
            ),
            "setval": "set forward-class"
            "{{ (' ' + set.forward_class|string) if set.forward_class is defined else '' }}",
            "result": {
                "policies": {
                    "set": {
                        "forward_class": "{{ forward_class }}",
                    },
                },
            },
        },
        {
            "name": "set.ip_precedence",
            "getval": re.compile(
                r"""
                \s*set\sip-precedence
                (\s(?P<ip_precedence>\d+))
                $""", re.VERBOSE,
            ),
            "setval": "set ip-precedence"
            "{{ (' ' + set.ip_precedence|string) if set.ip_precedence is defined else '' }}",
            "result": {
                "policies": {
                    "set": {
                        "ip_precedence": "{{ ip_precedence }}",
                    },
                },
            },
        },
        {
            "name": "set.isis_metric",
            "getval": re.compile(
                r"""
                \s*set\sisis-metric
                (\s(?P<isis_metric>\d+))
                $""", re.VERBOSE,
            ),
            "setval": "set isis-metric"
            "{{ (' ' + set.isis_metric|string) if set.isis_metric is defined else '' }}",
            "result": {
                "policies": {
                    "set": {
                        "isis_metric": "{{ isis_metric }}",
                    },
                },
            },
        },
        {
            "name": "set.label",
            "getval": re.compile(
                r"""
                \s*set\slabel
                (\s(?P<label>\d+))
                $""", re.VERBOSE,
            ),
            "setval": "set label"
            "{{ (' ' + set.label|string) if set.label is defined else '' }}",
            "result": {
                "policies": {
                    "set": {
                        "label": "{{ label }}",
                    },
                },
            },
        },
        {
            "name": "set.label_index",
            "getval": re.compile(
                r"""
                \s*set\slabel-index
                (\s(?P<label_index>\d+))
                $""", re.VERBOSE,
            ),
            "setval": "set label-index"
            "{{ (' ' + set.label_index|string) if set.label_index is defined else '' }}",
            "result": {
                "policies": {
                    "set": {
                        "label_index": "{{ label_index }}",
                    },
                },
            },
        },
        {
            "name": "set.label_mode",
            "getval": re.compile(
                r"""
                \s*set\slabel-mode
                (\s(?P<per_ce>per-ce))?
                (\s(?P<per_prefix>per-prefix))?
                (\s(?P<per_vrf>per-vrf))?
                $""", re.VERBOSE,
            ),
            "setval": "set label-mode"
            "{{ (' per-ce' ) if set.label_mode.per_ce|d(False) is defined else '' }}"
            "{{ (' per-prefix' ) if set.label_mode.per_prefix|d(False) is defined else '' }}"
            "{{ (' per-vrf' ) if set.label_mode.per_vrf|d(False) is defined else '' }}",
            "result": {
                "policies": {
                    "set": {
                        "label_mode": {
                            "per_ce": "{{ not not per_ce }}",
                            "per_prefix": "{{ not not per_prefix }}",
                            "per_vrf": "{{ not not per_vrf }}",
                        },
                    },
                },
            },
        },
        {
            "name": "set.large_community",
            "getval": re.compile(
                r"""
                \s*set\slarge-community
                (\s(?P<large_community>.+))
                $""", re.VERBOSE,
            ),
            "setval": "set large-community {{ set.large_community }}",
            "result": {
                "policies": {
                    "set": {
                        "large_community": "{{ large_community }}",
                    },
                },
            },
        },
        {
            "name": "set.level",
            "getval": re.compile(
                r"""
                \s*set\slevel
                (\s(?P<level_1>level-1))?
                (\s(?P<level_1_2>level-1-2))?
                (\s(?P<level_2>level-2))?
                $""", re.VERBOSE,
            ),
            "setval": "set level"
            "{{ (' level-1' ) if set.level.level_1|d(False) is defined else '' }}"
            "{{ (' level-1-2' ) if set.level.level_1_2|d(False) is defined else '' }}"
            "{{ (' level-2' ) if set.level.level_2|d(False) is defined else '' }}",
            "result": {
                "policies": {
                    "set": {
                        "level": {
                            "level_1": "{{ not not level_1 }}",
                            "level_1_2": "{{ not not level_1_2 }}",
                            "level_2": "{{ not not level_2 }}",
                        },
                    },
                },
            },
        },
        {
            "name": "set.load_balance",
            "getval": re.compile(
                r"""
                \s*set\sload-balance\secmp-consistent
                $""", re.VERBOSE,
            ),
            "setval": "set load-balance ecmp-consistent",
            "result": {
                "policies": {
                    "set": {
                        "load_balance": True,
                    },
                },
            },
        },
        {
            "name": "set.lsm_root",
            "getval": re.compile(
                r"""
                \s*set\slsm-root
                (\s(?P<lsm_root>\S+))
                $""", re.VERBOSE,
            ),
            "setval": "set lsm-root {{ set.lsm_root }}",
            "result": {
                "policies": {
                    "set": {
                        "lsm_root": "{{ lsm_root }}",
                    },
                },
            },
        },
        {
            "name": "set.metric_type",
            "getval": re.compile(
                r"""
                \s*set\smetric-type
                (\s(?P<external>external))?
                (\s(?P<internal>internal))?
                (\s(?P<rib_metric_as_external>rib-metric-as-external))?
                (\s(?P<rib_metric_as_internal>rib-metric-as-internal))?
                (\s(?P<type_1>type-1))?
                (\s(?P<type_2>type-2))?
                $""", re.VERBOSE,
            ),
            "setval": "set metric-type"
            "{{ (' external' ) if set.metric_type.external|d(False) is defined else '' }}"
            "{{ (' internal' ) if set.metric_type.internal|d(False) is defined else '' }}"
            "{{ (' rib-metric-as-external' ) if set.metric_type.rib_metric_as_external|d(False) is defined else '' }}"
            "{{ (' rib-metric-as-internal' ) if set.metric_type.rib_metric_as_internal|d(False) is defined else '' }}"
            "{{ (' type-1' ) if set.metric_type.type_1|d(False) is defined else '' }}"
            "{{ (' type-2' ) if set.metric_type.type_2|d(False) is defined else '' }}",
            "result": {
                "policies": {
                    "set": {
                        "metric_type": {
                            "external": "{{ not not external }}",
                            "internal": "{{ not not internal }}",
                            "rib_metric_as_external": "{{ not not rib_metric_as_external }}",
                            "rib_metric_as_internal": "{{ not not rib_metric_as_internal }}",
                            "type_1": "{{ not not type_1 }}",
                            "type_2": "{{ not not type_2 }}",
                        },
                    },
                },
            },
        },
        {
            "name": "set.mpls",
            "getval": re.compile(
                r"""
                \s*set\smpls\straffic-eng\sattributeset
                (\s(?P<mpls>\S+))
                $""", re.VERBOSE,
            ),
            "setval": "set mpls traffic-eng attributeset {{ set.mpls }}",
            "result": {
                "policies": {
                    "set": {
                        "mpls": "{{ mpls }}",
                    },
                },
            },
        },
        {
            "name": "set.med",
            "getval": re.compile(
                r"""
                \s*set\smed
                (\s(?P<increment>\+))?
                (\s(?P<decrement>\-))?
                (\s(?P<value>\d+))?
                (\s(?P<igp_cost>igp-cost))?
                (\s(?P<max_reachable>max-reachable))?
                (\s(?P<parameter>\$\w+))?
                $""", re.VERBOSE,
            ),
            "setval": "set med"
            "{{ (' +' ) if set.med.increment is defined else '' }}"
            "{{ (' -' ) if set.med.decrement is defined else '' }}"
            "{{ (' ' + set.med.value|string ) if set.med.value is defined else '' }}"
            "{{ (' igp-cost') if set.med.igp_cost|d(False) else '' }}"
            "{{ (' max-reachable') if set.med.max_reachable|d(False) else '' }}"
            "{{ (' ' + set.med.parameter ) if set.med.parameter is defined else '' }}",
            "result": {
                "policies": {
                    "set": {
                        "med": {
                            "increment": "{{ not not increment }}",
                            "decrement": "{{ not not decrement }}",
                            "value": "{{ value }}",
                            "igp_cost": "{{ not not igp_cost }}",
                            "max_reachable": "{{ not not max_reachable }}",
                            "parameter": "{{ parameter }}",
                        },
                    },
                },
            },
        },
        {
            "name": "set.next_hop",
            "getval": re.compile(
                r"""
                \s*set\snext-hop
                (\s(?P<next_hop>\S+))
                $""", re.VERBOSE,
            ),
            "setval": "set next-hop {{ set.next_hop.address }}",
            "result": {
                "policies": {
                    "set": {
                        "next_hop": {
                            "address": "{{ next_hop }}",
                        },
                    },
                },
            },
        },
        {
            "name": "set.origin",
            "getval": re.compile(
                r"""
                \s*set\sorigin
                (\s(?P<egp>egp))?
                (\s(?P<igp>igp))?
                (\s(?P<rincomplete>incomplete))?
                $""", re.VERBOSE,
            ),
            "setval": "set origin"
            "{{ (' egp' ) if set.origin.egp|d(False) is defined else '' }}"
            "{{ (' igp' ) if set.origin.igp|d(False) is defined else '' }}"
            "{{ (' incomplete' ) if set.origin.incomplete|d(False) is defined else '' }}",
            "result": {
                "policies": {
                    "set": {
                        "origin": {
                            "egp": "{{ not not egp }}",
                            "igp": "{{ not not igp }}",
                            "rincomplete": "{{ not not rincomplete }}",
                        },
                    },
                },
            },
        },
        {
            "name": "set.ospf_metric",
            "getval": re.compile(
                r"""
                \s*set\sospf-metric
                (\s(?P<ospf_metric>\d+))
                $""", re.VERBOSE,
            ),
            "setval": "set ospf-metric {{ ospf_metric|string }}",
            "result": {
                "policies": {
                    "set": {
                        "ospf_metric": "{{ ospf_metric }}",
                    },
                },
            },
        },
        {
            "name": "set.path_selection.all",
            "getval": re.compile(
                r"""
                \s*set\spath-selection\sall\sadvertise
                $""", re.VERBOSE,
            ),
            "setval": "set path-selection all advertise",
            "result": {
                "policies": {
                    "set": {
                        "path_selection": {
                            "all": True,
                        },
                    },
                },
            },
        },
        {
            "name": "set.path_selection.backup",
            "getval": re.compile(
                r"""
                \s*set\spath-selection\sbackup
                (\s(?P<backup_decimal>\d+))?
                (\s(?P<advertise>advertise))?
                (\s(?P<install>install))?
                $""", re.VERBOSE,
            ),
            "setval": "set path-selection backup"
            "{{ (' ' + set.path_selection.backup.backup_decimal|string ) if set.path_selection.backup.backup_decimal is defined else '' }}"
            "{{ (' advertise') if set.path_selection.backup.advertise|d(False) else '' }}"
            "{{ (' install') if set.path_selection.backup.install|d(False) else '' }}",
            "result": {
                "policies": {
                    "set": {
                        "path_selection": {
                            "backup": {
                                "backup_decimal": "{{ backup_decimal }}",
                                "advertise": "{{ not not advertise }}",
                                "install": "{{ not not install }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "set.path_selection.best_path",
            "getval": re.compile(
                r"""
                \s*set\spath-selection\sbest-path
                $""", re.VERBOSE,
            ),
            "setval": "set path-selection best_path",
            "result": {
                "policies": {
                    "set": {
                        "path_selection": {
                            "best_path": True,
                        },
                    },
                },
            },
        },
        {
            "name": "set.path_selection.group_best",
            "getval": re.compile(
                r"""
                \s*set\spath-selection\sbest-path\sadvertise
                $""", re.VERBOSE,
            ),
            "setval": "set path-selection group-best advertise",
            "result": {
                "policies": {
                    "set": {
                        "path_selection": {
                            "group_best": True,
                        },
                    },
                },
            },
        },
        {
            "name": "set.path_selection.multiplath",
            "getval": re.compile(
                r"""
                \s*set\spath-selection\smultiplath\sadvertise
                $""", re.VERBOSE,
            ),
            "setval": "set path-selection multiplath advertise",
            "result": {
                "policies": {
                    "set": {
                        "path_selection": {
                            "multiplath": True,
                        },
                    },
                },
            },
        },
        {
            "name": "set.path_color",
            "getval": re.compile(
                r"""
                \s*set\spath-color\sexternal-reach
                $""", re.VERBOSE,
            ),
            "setval": "set path-color external-reach",
            "result": {
                "policies": {
                    "set": {
                        "path_color": True,
                    },
                },
            },
        },
        {
            "name": "set.qos_group",
            "getval": re.compile(
                r"""
                \s*set\sqos-group
                (\s(?P<qos_group>\d+))
                $""", re.VERBOSE,
            ),
            "setval": "set qos-group {{ qos_group|string }}",
            "result": {
                "policies": {
                    "set": {
                        "qos_group": "{{ qos_group }}",
                    },
                },
            },
        },
        {
            "name": "set.rib_metric",
            "getval": re.compile(
                r"""
                \s*set\srib-metric
                (\s(?P<rib_metric>\d+))
                $""", re.VERBOSE,
            ),
            "setval": "set rib-metric {{ rib_metric|string }}",
            "result": {
                "policies": {
                    "set": {
                        "rib_metric": "{{ rib_metric }}",
                    },
                },
            },
        },
        {
            "name": "set.rip_metric",
            "getval": re.compile(
                r"""
                \s*set\srip-metric
                (\s(?P<rip_metric>\d+))
                $""", re.VERBOSE,
            ),
            "setval": "set rip-metric {{ rip_metric|string }}",
            "result": {
                "policies": {
                    "set": {
                        "rip_metric": "{{ rip_metric }}",
                    },
                },
            },
        },
        {
            "name": "set.rip_tag",
            "getval": re.compile(
                r"""
                \s*set\srip-tag
                (\s(?P<rip_tag>\d+))
                $""", re.VERBOSE,
            ),
            "setval": "set rip-tag {{ rip_tag|string }}",
            "result": {
                "policies": {
                    "set": {
                        "rip_tag": "{{ rip_tag }}",
                    },
                },
            },
        },
        {
            "name": "set.rt_set",
            "getval": re.compile(
                r"""
                \s*set\srt-set\sroute-limit
                (\s(?P<rt_set>\d+))
                $""", re.VERBOSE,
            ),
            "setval": "set rt-set route-limit {{ rt_set|string }}",
            "result": {
                "policies": {
                    "set": {
                        "rt_set": "{{ rt_set }}",
                    },
                },
            },
        },
        {
            "name": "set.s_pmsi",
            "getval": re.compile(
                r"""
                \s*set\ss-pmsi\sstar-g
                $""", re.VERBOSE,
            ),
            "setval": "set s-pmsi star-g",
            "result": {
                "policies": {
                    "set": {
                        "s_pmsi": True,
                    },
                },
            },
        },
        {
            "name": "set.spf_priority",
            "getval": re.compile(
                r"""
                \s*set\sspf-priority
                (\s(?P<critical>critical))?
                (\s(?P<high>high))?
                (\s(?P<medium>medium))?
                $""", re.VERBOSE,
            ),
            "setval": "set spf-priority"
            "{{ (' critical' ) if set.spf_priority.critical|d(False) is defined else '' }}"
            "{{ (' high' ) if set.spf_priority.high|d(False) is defined else '' }}"
            "{{ (' medium' ) if set.spf_priority.medium|d(False) is defined else '' }}",
            "result": {
                "policies": {
                    "set": {
                        "spf_priority": {
                            "critical": "{{ not not critical }}",
                            "high": "{{ not not high }}",
                            "medium": "{{ not not medium }}",
                        },
                    },
                },
            },
        },
        {
            "name": "set.static_p2mp_te",
            "getval": re.compile(
                r"""
                \s*set\sstatic-p2mp-te
                (\s(?P<static_p2mp_te>\S+))
                $""", re.VERBOSE,
            ),
            "setval": "set static-p2mp-te {{ set.static_p2mp_te }}",
            "result": {
                "policies": {
                    "set": {
                        "static_p2mp_te": "{{ static_p2mp_te }}",
                    },
                },
            },
        },
        {
            "name": "set.tag",
            "getval": re.compile(
                r"""
                \s*set\stag
                (\s(?P<tag>\d+))
                $""", re.VERBOSE,
            ),
            "setval": "set tag {{ set.tag|string }}",
            "result": {
                "policies": {
                    "set": {
                        "tag": "{{ tag }}",
                    },
                },
            },
        },
        {
            "name": "set.traffic_index",
            "getval": re.compile(
                r"""
                \s*set\straffic-index
                (\s(?P<index_number>\d+))?
                (\s(?P<ignore>ignore))?
                $""", re.VERBOSE,
            ),
            "setval": "set traffic-index"
            "{{ (' ' + set.traffic_index.index_number|string) if set.traffic_index.index_number is defined else '' }}"
            "{{ (' ignore') if set.traffic_index.ignore is defined else '' }}",
            "result": {
                "policies": {
                    "set": {
                        "traffic_index": {
                            "index_number": "{{ index_number }}",
                            "ignore": "{{ not not ignore }}",
                        },
                    },
                },
            },
        },
        {
            "name": "set.upstream_core_tree",
            "getval": re.compile(
                r"""
                \s*set\supstream-core-tree
                (\s(?P<ingress_replication>ingress-replication))?
                (\s(?P<mldp>mldp))?
                (\s(?P<p2mp_te>p2mp-te))?
                (\s(?P<sr_p2mp>sr-p2mp))?
                $""", re.VERBOSE,
            ),
            "setval": "set upstream-core-tree"
            "{{ (' ingress-replication' ) if set.upstream_core_tree.ingress_replication|d(False) is defined else '' }}"
            "{{ (' mldp' ) if set.upstream_core_tree.mldp|d(False) is defined else '' }}"
            "{{ (' p2mp-te' ) if set.upstream_core_tree.p2mp_te|d(False) is defined else '' }}"
            "{{ (' sr-p2mp' ) if set.upstream_core_tree.sr_p2mp|d(False) is defined else '' }}",
            "result": {
                "policies": {
                    "set": {
                        "upstream_core_tree": {
                            "ingress_replication": "{{ not not ingress_replication }}",
                            "mldp": "{{ not not mldp }}",
                            "p2mp_te": "{{ not not p2mp_te }}",
                            "sr_p2mp": "{{ not not sr_p2mp }}",
                        },
                    },
                },
            },
        },
        {
            "name": "set.vpn_distinguisher",
            "getval": re.compile(
                r"""
                \s*set\svpn-distinguisher
                (\s(?P<vpn_distinguisher>\d+))
                $""", re.VERBOSE,
            ),
            "setval": "set vpn-distinguisher {{ set.vpn_distinguisher|string }}",
            "result": {
                "policies": {
                    "set": {
                        "vpn_distinguisher": "{{ vpn_distinguisher }}",
                    },
                },
            },
        },
        {
            "name": "set.weight",
            "getval": re.compile(
                r"""
                \s*set\sweight
                (\s(?P<weight>\d+))
                $""", re.VERBOSE,
            ),
            "setval": "set weight {{ set.weight|string }}",
            "result": {
                "policies": {
                    "set": {
                        "weight": "{{ weight }}",
                    },
                },
            },
        },
    ]
    # fmt: on



# -*- coding: utf-8 -*-

# Copyright: Contributors to the Ansible project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

try:
    import boto3
except ImportError:
    pass
from typing import Any
from typing import Dict
from typing import Optional
from typing import Tuple


def generate_encryption_settings(
    bucket_sse_mode: Optional[str], bucket_sse_kms_key_id: Optional[str]
) -> Tuple[Dict, Dict]:
    """Generate Encryption Settings"""
    args = {}
    headers = {}
    if not bucket_sse_mode:
        return args, headers

    args["ServerSideEncryption"] = bucket_sse_mode
    headers["x-amz-server-side-encryption"] = bucket_sse_mode
    if bucket_sse_mode == "aws:kms" and bucket_sse_kms_key_id:
        args["SSEKMSKeyId"] = bucket_sse_kms_key_id
        headers["x-amz-server-side-encryption-aws-kms-key-id"] = bucket_sse_kms_key_id
    return args, headers


class S3ClientManager:
    def __init__(self, client: Any) -> None:
        self._s3_client = client

    @property
    def client(self) -> Any:
        return self._s3_client

    @staticmethod
    def _get_s3_client(
        access_key_id: Optional[str],
        secret_key_id: Optional[str],
        session_token: Optional[str],
        region_name: str,
        profile_name: Optional[str],
    ) -> Any:
        if not region_name:
            region_name = "us-east-1"
        session_args = dict(
            aws_access_key_id=access_key_id,
            aws_secret_access_key=secret_key_id,
            aws_session_token=session_token,
            region_name=region_name,
        )
        if profile_name:
            session_args["profile_name"] = profile_name
        session = boto3.session.Session(**session_args)
        return session.client("s3")

    @staticmethod
    def get_bucket_endpoint(
        bucket_name: str,
        bucket_endpoint_url: Optional[str],
        access_key_id: Optional[str],
        secret_key_id: Optional[str],
        session_token: Optional[str],
        region_name: Optional[str],
        profile_name: Optional[str],
    ) -> Tuple[str, str]:
        """
        Fetches the correct S3 endpoint and region for use with our bucket.
        If we don't explicitly set the endpoint then some commands will use the global
        endpoint and fail
        (new AWS regions and new buckets in a region other than the one we're running in)
        """
        if not region_name:
            region_name = "us-east-1"
        tmp_s3_client = S3ClientManager._get_s3_client(
            access_key_id, secret_key_id, session_token, region_name, profile_name
        )

        # Fetch the location of the bucket so we can open a client against the 'right' endpoint
        # This /should/ always work
        head_bucket = tmp_s3_client.head_bucket(Bucket=(bucket_name))
        bucket_region = head_bucket.get("ResponseMetadata", {}).get("HTTPHeaders", {}).get("x-amz-bucket-region", None)
        if bucket_region is None:
            bucket_region = "us-east-1"

        if bucket_endpoint_url:
            return bucket_endpoint_url, bucket_region

        # Create another client for the region the bucket lives in, so we can nab the endpoint URL
        if bucket_region != region_name:
            tmp_s3_client = S3ClientManager._get_s3_client(
                access_key_id, secret_key_id, session_token, bucket_region, profile_name
            )

        return tmp_s3_client.meta.endpoint_url, tmp_s3_client.meta.region_name

    def get_url(
        self,
        client_method: str,
        bucket_name: str,
        out_path: str,
        http_method: str,
        extra_args: Optional[Dict[str, Any]] = None,
    ) -> str:
        """Generate URL for get_object / put_object"""

        params = {"Bucket": bucket_name, "Key": out_path}
        if extra_args is not None:
            params.update(extra_args)
        return self._s3_client.generate_presigned_url(
            client_method, Params=params, ExpiresIn=3600, HttpMethod=http_method
        )

    def generate_host_commands(
        self,
        bucket_name: str,
        bucket_sse_mode: Optional[str],
        bucket_sse_kms_key_id: Optional[str],
        s3_path: str,
        in_path: str,
        out_path: str,
        is_windows: bool,
        method: str,
    ) -> Tuple[str, Optional[Dict[str, str]]]:
        """
        Generate commands for the specified bucket and configurations, S3 path, input path, and output path.
        The function generates a curl/Invoke-WebRequest command to be executed on the managed node.
        The method 'get' means the controller node wants to retrieve content from the managed node, the corresponding
        on the managed node is a 'put' command to updload content to the S3 bucket.
        The method 'put' means the controller node wants to upload content to the managed node, the corresponding
        on the managed node is a 'get' command to download content to the S3 bucket.

        :param bucket_name: The name of the S3 bucket used for file transfers.
        :param s3_path: The S3 path to the file to be sent.
        :param in_path: Input path
        :param out_path: Output path
        :param method: The request method to use for the command (can be "get" or "put").

        :returns: A tuple containing a shell command string and optional extra arguments for the PUT request.
        """

        command = None
        url = None
        put_args = None
        if method == "get":
            put_args, put_headers = generate_encryption_settings(bucket_sse_mode, bucket_sse_kms_key_id)
            url = self.get_url("put_object", bucket_name, s3_path, "PUT", extra_args=put_args)
            if is_windows:
                put_command_headers = "; ".join([f"'{h}' = '{v}'" for h, v in put_headers.items()])
                command = (
                    "Invoke-WebRequest -Method PUT "
                    # @{'key' = 'value'; 'key2' = 'value2'}
                    f"-Headers @{{{put_command_headers}}} "
                    f"-InFile '{in_path}' "
                    f"-Uri '{url}' "
                    f"-UseBasicParsing"
                )  # fmt: skip
            else:
                # Due to https://github.com/curl/curl/issues/183 earlier
                # versions of curl did not create the output file, when the
                # response was empty. Although this issue was fixed in 2015,
                # some actively maintained operating systems still use older
                # versions of it (e.g. CentOS 7)
                put_command_headers = " ".join([f"-H '{h}: {v}'" for h, v in put_headers.items()])
                command = (
                    "curl --request PUT "
                    f"{put_command_headers} "
                    f"--upload-file '{in_path}' "
                    f"'{url}'"
                )  # fmt: skip
        elif method == "put":
            url = self.get_url("get_object", bucket_name, s3_path, "GET")
            if is_windows:
                command = (
                    "Invoke-WebRequest "
                    f"'{url}' "
                    f"-OutFile '{out_path}'"
                )  # fmt: skip
            else:
                command = (
                    "curl "
                    f"-o '{out_path}' "
                    f"'{url}'"
                    ";"
                    "touch "
                    f"'{out_path}'"
                )  # fmt: skip

        return command, put_args



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or
# https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
module: networks_wireless_settings
short_description: Resource module for networks _wireless _settings
description:
  - Manage operation update of the resource networks _wireless _settings.
  - Update the wireless settings for a network.
version_added: '1.0.0'
extends_documentation_fragment:
  - cisco.meraki.module
author: Francisco Munoz (@fmunoz)
options:
  ipv6BridgeEnabled:
    description: Toggle for enabling or disabling IPv6 bridging in a network (Note
      if enabled, SSIDs must also be configured to use bridge mode).
    type: bool
  ledLightsOn:
    description: Toggle for enabling or disabling LED lights on all APs in the network
      (making them run dark).
    type: bool
  locationAnalyticsEnabled:
    description: Toggle for enabling or disabling location analytics for your network.
    type: bool
  meshingEnabled:
    description: Toggle for enabling or disabling meshing in a network.
    type: bool
  multicastToUnicastConversion:
    description: Multicast-to-unicast conversion settings across the network.
    suboptions:
      enabled:
        description: Toggle for enabling or disabling multicast-to-unicast conversion
          across the network.
        type: bool
    type: dict
  namedVlans:
    description: Named VLAN settings for wireless networks.
    suboptions:
      poolDhcpMonitoring:
        description: Named VLAN Pool DHCP Monitoring settings.
        suboptions:
          duration:
            description: The duration in minutes that devices will refrain from using
              dirty VLANs before adding them back to the pool.
            type: int
          enabled:
            description: Whether or not devices using named VLAN pools should remove
              dirty VLANs from the pool, thereby preventing clients from being assigned
              to VLANs where they would be unable to obtain an IP address via DHCP.
            type: bool
        type: dict
    type: dict
  networkId:
    description: NetworkId path parameter. Network ID.
    type: str
  upgrade:
    description: Upgrade settings for the network.
    suboptions:
      predownload:
        description: Predownload settings for firmware upgrades.
        suboptions:
          enabled:
            description: The default setting for upgrade firmware predownload. Only
              applies to devices running MR 32 or higher.
            type: bool
        type: dict
    type: dict
  upgradeStrategy:
    description: The default strategy that network devices will use to perform an
      upgrade. Requires firmware version MR 26.8 or higher.
    type: str
requirements:
  - meraki >= 2.4.9
  - python >= 3.5
seealso:
  - name: Cisco Meraki documentation for wireless updateNetworkWirelessSettings
    description: Complete reference of the updateNetworkWirelessSettings API.
    link: https://developer.cisco.com/meraki/api-v1/#!update-network-wireless-settings
notes:
  - SDK Method used are
    wireless.Wireless.update_network_wireless_settings,
  - Paths used are
    put /networks/{networkId}/wireless/settings,
"""

EXAMPLES = r"""
- name: Update all
  cisco.meraki.networks_wireless_settings:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    state: present
    ipv6BridgeEnabled: false
    ledLightsOn: false
    locationAnalyticsEnabled: false
    meshingEnabled: true
    multicastToUnicastConversion:
      enabled: true
    namedVlans:
      poolDhcpMonitoring:
        duration: 5
        enabled: true
    networkId: string
    upgrade:
      predownload:
        enabled: false
    upgradeStrategy: minimizeUpgradeTime
"""
RETURN = r"""
meraki_response:
  description: A dictionary or list with the response returned by the Cisco Meraki Python SDK
  returned: always
  type: dict
  sample: >
    {
      "meshingEnabled": true,
      "ipv6BridgeEnabled": true,
      "locationAnalyticsEnabled": true,
      "upgradeStrategy": "string",
      "upgrade": {
        "predownload": {
          "enabled": true
        }
      },
      "ledLightsOn": true,
      "multicastToUnicastConversion": {
        "enabled": true
      },
      "namedVlans": {
        "poolDhcpMonitoring": {
          "enabled": true,
          "duration": 0
        }
      },
      "regulatoryDomain": {
        "name": "string",
        "countryCode": "string",
        "permits6e": true
      }
    }
"""



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2024, Akini Ross (@akinross) <akinross@cisco.com>
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {"metadata_version": "1.1", "status": ["preview"], "supported_by": "certified"}

DOCUMENTATION = r"""
---
module: aci_fabric_management_access
short_description: Manage Fabric Management Access (comm:Pol)
description:
- Manage Fabric Management Access on Cisco ACI fabrics.
options:
  name:
    description:
    - The name of the Fabric Management Access policy.
    type: str
    aliases: [ fabric_management_access_policy_name ]
  description:
    description:
    - The description of the Fabric Management Access policy.
    type: str
    aliases: [ descr ]
  name_alias:
    description:
    - The name alias of the Fabric Management Access policy.
    - This relates to the nameAlias property in ACI.
    type: str
  http:
    description:
    - Parameters for HTTP configuration (comm:Http).
    type: dict
    suboptions:
      admin_state:
        description:
        - The admin state of the HTTP connection.
        - The APIC defaults to C(disabled) when unset during creation.
        type: str
        choices: [ enabled, disabled ]
      port:
        description:
        - The port for the HTTP connection.
        - The APIC defaults to C(80) when unset during creation.
        type: int
      redirect:
        description:
        - The state of the HTTP to HTTPS redirect service.
        - The APIC defaults to C(disabled) when unset during creation.
        type: str
        choices: [ enabled, disabled, tested ]
      allow_origins:
        description:
        - The allowed origins for the HTTP connection.
        - 'Example format: http://127.0.0.1:8000'
        type: str
      allow_credentials:
        description:
        - The state of the allow credential for the HTTP connection.
        - The APIC defaults to C(disabled) when unset during creation.
        type: str
        choices: [ enabled, disabled ]
      throttle:
        description:
        - The state of the request throttle for the HTTP connection.
        - The APIC defaults to C(disabled) when unset during creation.
        type: str
        choices: [ enabled, disabled ]
      throttle_rate:
        description:
        - The rate of the request throttle.
        - The APIC defaults to C(10000) when unset during creation.
        type: int
      throttle_unit:
        description:
        - The unit of the request throttle rate.
        - The APIC defaults to C(requests_per_second) when unset during creation.
        type: str
        choices: [ requests_per_second, requests_per_minute ]
  https:
    description:
    - Parameters for HTTPS configuration (comm:Https).
    type: dict
    suboptions:
      admin_state:
        description:
        - The admin state of the HTTPS connection.
        - The APIC defaults to C(enabled) when unset during creation.
        type: str
        choices: [ enabled, disabled ]
      port:
        description:
        - The port for the HTTPS connection.
        - The APIC defaults to C(443) when unset during creation.
        type: int
      allow_origins:
        description:
        - The allowed origins for the HTTPS connection.
        - 'Example format: http://127.0.0.1:8000'
        type: str
      allow_credentials:
        description:
        - The state of the allow credential for the HTTPS connection.
        - The APIC defaults to C(disabled) when unset during creation.
        type: str
        choices: [ enabled, disabled ]
      ssl:
        description:
        - The SSL protocol(s) for the HTTPS connection.
        - The APIC defaults to C(tls_v1.1) and C(tls_v1.2) set when unset during creation.
        type: list
        elements: str
        choices: [ tls_v1.0, tls_v1.1, tls_v1.2, tls_v1.3 ]
        aliases: [ ssl_protocols ]
      dh_param:
        description:
        - The Diffie-Hellman parameter for the HTTPS connection.
        - The APIC defaults to C(none) when unset during creation.
        type: str
        choices: [ '1024', '2048', '4096', none ]
      throttle:
        description:
        - The state of the request throttle for the HTTPS connection.
        - The APIC defaults to C(disabled) when unset during creation.
        type: str
        choices: [ enabled, disabled ]
      throttle_rate:
        description:
        - The rate of the request throttle.
        - The APIC defaults to C(10000) when unset during creation.
        type: int
      throttle_unit:
        description:
        - The unit of the request throttle rate.
        - The APIC defaults to C(requests_per_second) when unset during creation.
        type: str
        choices: [ requests_per_second, requests_per_minute ]
      admin_key_ring:
        description:
        - The admin key ring for the HTTPS connection.
        - The APIC defaults to C(default) when unset during creation.
        type: str
      client_certificate_trustpoint:
        description:
        - The client certificate trustpoint for the HTTPS connection.
        type: str
        aliases: [ trustpoint ]
      client_certificate_authentication_state:
        description:
        - The client certificate authentication state for the HTTPS connection.
        - The APIC defaults to C(disabled) when unset during creation.
        - The C(enabled) state requires a C(client_certificate_trustpoint) to be set.
        type: str
        choices: [ enabled, disabled ]
        aliases: [ client_certificate_auth_state, auth_state, authentication_state ]
  telnet:
    description:
    - Parameters for telnet configuration (comm:Telnet).
    type: dict
    suboptions:
      admin_state:
        description:
        - The admin state of the telnet connection.
        - The APIC defaults to C(disabled) when unset during creation.
        type: str
        choices: [ enabled, disabled ]
      port:
        description:
        - The port for the telnet connection.
        - The APIC defaults to C(23) when unset during creation.
        type: int
  ssh:
    description:
    - Parameters for SSH configuration (comm:Ssh).
    type: dict
    suboptions:
      admin_state:
        description:
        - The admin state of the SSH connection.
        - The APIC defaults to C(enabled) when unset during creation.
        type: str
        choices: [ enabled, disabled ]
      password_auth_state:
        description:
        - The password authentication state of the SSH connection.
        - The APIC defaults to C(enabled) when unset during creation.
        type: str
        choices: [ enabled, disabled ]
      port:
        description:
        - The port for the SSH connection.
        - The APIC defaults to C(22) when unset during creation.
        type: int
      ciphers:
        description:
        - The ciphers of the SSH connection.
        - The APIC defaults to all options set when unset during creation.
        type: list
        elements: str
        choices: [ aes128_ctr, aes192_ctr, aes256_ctr, aes128_gcm, aes256_gcm, chacha20 ]
      kex:
        description:
        - The KEX algorithms of the SSH connection.
        - The APIC defaults to all options set when unset during creation.
        type: list
        elements: str
        choices: [ dh_sha1, dh_sha256, dh_sha512, curve_sha256, curve_sha256_libssh, ecdh_256, ecdh_384, ecdh_521 ]
      macs:
        description:
        - The MACs of the SSH connection.
        - The APIC defaults to all options set  when unset during creation.
        type: list
        elements: str
        choices: [ sha1, sha2_256, sha2_512, sha2_256_etm, sha2_512_etm ]
  ssh_web:
    description:
    - Parameters for SSH access via WEB configuration (comm:Shellinabox).
    type: dict
    suboptions:
      admin_state:
        description:
        - The admin state of the SSH access via WEB connection.
        - The APIC defaults to C(disabled) when unset during creation.
        type: str
        choices: [ enabled, disabled ]
  state:
    description:
    - Use C(present) for updating configuration.
    - Use C(query) for showing current configuration.
    type: str
    choices: [ absent, present, query ]
    default: present
extends_documentation_fragment:
- cisco.aci.aci
- cisco.aci.annotation
- cisco.aci.owner

seealso:
- name: APIC Management Information Model reference
  description: More information about the internal APIC class B(comm:Pol).
  link: https://developer.cisco.com/docs/apic-mim-ref/
author:
- Akini Ross (@akinross)
"""

EXAMPLES = r"""
- name: Create a Fabric Management Access policy
  cisco.aci.aci_fabric_management_access:
    host: apic
    username: admin
    password: SomeSecretPassword
    name: fabric_management_access_policy_1
    description: "This is a example Fabric Management Access policy."
    state: present
  delegate_to: localhost

- name: Create a Fabric Management Access policy with telnet enabled
  cisco.aci.aci_fabric_management_access:
    host: apic
    username: admin
    password: SomeSecretPassword
    name: fabric_management_access_policy_1
    description: "This is a example Fabric Management Access policy."
    telnet:
      admin_state: enabled
    state: present
  delegate_to: localhost

- name: Create a Fabric Management Access policy with SSH access via WEB enabled
  cisco.aci.aci_fabric_management_access:
    host: apic
    username: admin
    password: SomeSecretPassword
    name: fabric_management_access_policy_1
    description: "This is a example Fabric Management Access policy."
    ssh_web:
      admin_state: enabled
    state: present
  delegate_to: localhost

- name: Create a Fabric Management Access policy with SSH enabled and ciphers set
  cisco.aci.aci_fabric_management_access:
    host: apic
    username: admin
    password: SomeSecretPassword
    name: fabric_management_access_policy_1
    description: "This is a example Fabric Management Access policy."
    ssh:
      admin_state: enabled
      ciphers:
        - aes128_ctr
        - aes192_ctr
        - aes256_ctr
    state: present
  delegate_to: localhost

- name: Create a Fabric Management Access policy with HTTP enabled
  cisco.aci.aci_fabric_management_access:
    host: apic
    username: admin
    password: SomeSecretPassword
    name: fabric_management_access_policy_1
    description: "This is a example Fabric Management Access policy."
    http:
      admin_state: enabled
      allow_origins: http://127.0.0.1:8000
      throttle: enabled
      throttle_rate: 7500
      throttle_unit: requests_per_minute
    state: present
  delegate_to: localhost

- name: Create a Fabric Management Access policy with HTTPS enabled
  cisco.aci.aci_fabric_management_access:
    host: apic
    username: admin
    password: SomeSecretPassword
    name: fabric_management_access_policy_1
    description: "This is a example Fabric Management Access policy."
    https:
      admin_state: enabled
      port: 445
      allow_origins: http://127.0.0.1:8000
      allow_credentials: enabled
      ssl:
        - tls_v1.2
      dh_param: 4096
      throttle: enabled
      throttle_rate: 7500
      throttle_unit: requests_per_minute
      admin_key_ring: default
      client_certificate_trustpoint: ansible_trustpoint
      client_certificate_authentication_state: enabled
    state: present
  delegate_to: localhost

- name: Query a Fabric Management Access policy
  cisco.aci.aci_fabric_management_access:
    host: apic
    username: admin
    password: SomeSecretPassword
    name: fabric_management_access_policy_1
    state: query
  delegate_to: localhost
  register: query_result

- name: Query all Fabric Management Access policies
  cisco.aci.aci_fabric_management_access:
    host: apic
    username: admin
    password: SomeSecretPassword
    state: query
  delegate_to: localhost
  register: query_result

- name: Delete a Fabric Management Access policy
  cisco.aci.aci_fabric_management_access:
    host: apic
    username: admin
    password: SomeSecretPassword
    name: fabric_management_access_policy_1
    state: absent
  delegate_to: localhost
"""

RETURN = r"""
current:
  description: The existing configuration from the APIC after the module has finished
  returned: success
  type: list
  sample:
    [
        {
            "fvTenant": {
                "attributes": {
                    "descr": "Production environment",
                    "dn": "uni/tn-production",
                    "name": "production",
                    "nameAlias": "",
                    "ownerKey": "",
                    "ownerTag": ""
                }
            }
        }
    ]
error:
  description: The error information as returned from the APIC
  returned: failure
  type: dict
  sample:
    {
        "code": "122",
        "text": "unknown managed object class foo"
    }
raw:
  description: The raw output returned by the APIC REST API (xml or json)
  returned: parse error
  type: str
  sample: '<?xml version="1.0" encoding="UTF-8"?><imdata totalCount="1"><error code="122" text="unknown managed object class foo"/></imdata>'
sent:
  description: The actual/minimal configuration pushed to the APIC
  returned: info
  type: list
  sample:
    {
        "fvTenant": {
            "attributes": {
                "descr": "Production environment"
            }
        }
    }
previous:
  description: The original configuration from the APIC before the module has started
  returned: info
  type: list
  sample:
    [
        {
            "fvTenant": {
                "attributes": {
                    "descr": "Production",
                    "dn": "uni/tn-production",
                    "name": "production",
                    "nameAlias": "",
                    "ownerKey": "",
                    "ownerTag": ""
                }
            }
        }
    ]
proposed:
  description: The assembled configuration from the user-provided parameters
  returned: info
  type: dict
  sample:
    {
        "fvTenant": {
            "attributes": {
                "descr": "Production environment",
                "name": "production"
            }
        }
    }
filter_string:
  description: The filter string used for the request
  returned: failure or debug
  type: str
  sample: ?rsp-prop-include=config-only
method:
  description: The HTTP method used for the request to the APIC
  returned: failure or debug
  type: str
  sample: POST
response:
  description: The HTTP response from the APIC
  returned: failure or debug
  type: str
  sample: OK (30 bytes)
status:
  description: The HTTP status from the APIC
  returned: failure or debug
  type: int
  sample: 200
url:
  description: The HTTP url used for the request to the APIC
  returned: failure or debug
  type: str
  sample: https://10.11.12.13/api/mo/uni/tn-production.json
"""

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.aci.plugins.module_utils.aci import ACIModule, aci_argument_spec, aci_annotation_spec, aci_owner_spec
from ansible_collections.cisco.aci.plugins.module_utils.constants import THROTTLE_UNIT, SSH_CIPHERS, KEX_ALGORITHMS, SSH_MACS, HTTP_TLS_MAPPING


def main():
    argument_spec = aci_argument_spec()
    argument_spec.update(aci_annotation_spec())
    argument_spec.update(aci_owner_spec())
    argument_spec.update(
        name=dict(type="str", aliases=["fabric_management_access_policy_name"]),  # Not required for querying all objects
        description=dict(type="str", aliases=["descr"]),
        name_alias=dict(type="str"),
        http=dict(
            type="dict",
            options=dict(
                admin_state=dict(type="str", choices=["enabled", "disabled"]),
                port=dict(type="int"),
                redirect=dict(type="str", choices=["enabled", "disabled", "tested"]),
                allow_origins=dict(type="str"),
                allow_credentials=dict(type="str", choices=["enabled", "disabled"]),
                throttle=dict(type="str", choices=["enabled", "disabled"]),
                throttle_rate=dict(type="int"),
                throttle_unit=dict(type="str", choices=["requests_per_second", "requests_per_minute"]),
            ),
        ),
        https=dict(
            type="dict",
            options=dict(
                admin_state=dict(type="str", choices=["enabled", "disabled"]),
                port=dict(type="int"),
                allow_origins=dict(type="str"),
                allow_credentials=dict(type="str", choices=["enabled", "disabled"]),
                ssl=dict(
                    type="list",
                    elements="str",
                    choices=list(HTTP_TLS_MAPPING.keys()),
                    aliases=["ssl_protocols"],
                ),
                dh_param=dict(type="str", choices=["1024", "2048", "4096", "none"]),
                throttle=dict(type="str", choices=["enabled", "disabled"]),
                throttle_rate=dict(type="int"),
                throttle_unit=dict(type="str", choices=["requests_per_second", "requests_per_minute"]),
                admin_key_ring=dict(type="str", no_log=False),
                client_certificate_trustpoint=dict(type="str", aliases=["trustpoint"]),
                client_certificate_authentication_state=dict(
                    type="str",
                    choices=["enabled", "disabled"],
                    aliases=["client_certificate_auth_state", "auth_state", "authentication_state"],
                ),
            ),
        ),
        telnet=dict(
            type="dict",
            options=dict(
                admin_state=dict(type="str", choices=["enabled", "disabled"]),
                port=dict(type="int"),
            ),
        ),
        ssh=dict(
            type="dict",
            options=dict(
                admin_state=dict(type="str", choices=["enabled", "disabled"]),
                password_auth_state=dict(type="str", choices=["enabled", "disabled"]),
                port=dict(type="int"),
                ciphers=dict(type="list", elements="str", choices=list(SSH_CIPHERS.keys())),
                kex=dict(type="list", elements="str", choices=list(KEX_ALGORITHMS.keys())),
                macs=dict(type="list", elements="str", choices=list(SSH_MACS.keys())),
            ),
        ),
        ssh_web=dict(
            type="dict",
            options=dict(
                admin_state=dict(type="str", choices=["enabled", "disabled"]),
            ),
        ),
        state=dict(type="str", default="present", choices=["absent", "present", "query"]),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[
            ["state", "present", ["name"]],
            ["state", "absent", ["name"]],
        ],
    )

    aci = ACIModule(module)
    aci_class = "commPol"
    aci_child_classes = ["commSsh", "commHttp", "commHttps", "commTelnet", "commShellinabox"]

    name = module.params.get("name")
    description = module.params.get("description")
    name_alias = module.params.get("name_alias")
    http = module.params.get("http")
    https = module.params.get("https")
    telnet = module.params.get("telnet")
    ssh = module.params.get("ssh")
    ssh_web = module.params.get("ssh_web")
    state = module.params.get("state")

    aci.construct_url(
        root_class=dict(
            aci_class=aci_class,
            aci_rn="fabric/comm-{0}".format(name),
            module_object=name,
            target_filter={"name": name},
        ),
        child_classes=aci_child_classes,
    )

    aci.get_existing()

    if state == "present":
        child_configs = []

        if ssh:
            child_configs.append(
                dict(
                    commSsh=dict(
                        attributes=dict(
                            adminSt=ssh.get("admin_state"),
                            passwordAuth=ssh.get("password_auth_state"),
                            port=ssh.get("port"),
                            sshCiphers=",".join(sorted(SSH_CIPHERS.get(v) for v in set(ssh.get("ciphers")))) if ssh.get("ciphers") else None,
                            kexAlgos=",".join(sorted(KEX_ALGORITHMS.get(v) for v in set(ssh.get("kex")))) if ssh.get("kex") else None,
                            sshMacs=",".join(sorted(SSH_MACS.get(v) for v in set(ssh.get("macs")))) if ssh.get("macs") else None,
                        )
                    )
                )
            )

        if http:
            child_configs.append(
                dict(
                    commHttp=dict(
                        attributes=dict(
                            adminSt=http.get("admin_state"),
                            port=http.get("port"),
                            redirectSt=http.get("redirect"),
                            accessControlAllowOrigins=http.get("allow_origins"),
                            accessControlAllowCredential=http.get("allow_credentials"),
                            globalThrottleSt=http.get("throttle"),
                            globalThrottleRate=http.get("throttle_rate"),
                            globalThrottleUnit=THROTTLE_UNIT.get(http.get("throttle_unit")),
                        )
                    )
                )
            )

        if https:
            https_config = dict(
                commHttps=dict(
                    attributes=dict(
                        adminSt=https.get("admin_state"),
                        port=https.get("port"),
                        accessControlAllowOrigins=https.get("allow_origins"),
                        accessControlAllowCredential=https.get("allow_credentials"),
                        sslProtocols=",".join(sorted(HTTP_TLS_MAPPING.get(v) for v in set(https.get("ssl")))) if https.get("ssl") else None,
                        dhParam=https.get("dh_param"),
                        globalThrottleSt=https.get("throttle"),
                        globalThrottleRate=https.get("throttle_rate"),
                        globalThrottleUnit=THROTTLE_UNIT.get(https.get("throttle_unit")),
                        clientCertAuthState=https.get("client_certificate_authentication_state"),
                    ),
                    children=[],
                )
            )

            if https.get("admin_key_ring"):
                https_config["commHttps"]["children"].append(dict(commRsKeyRing=dict(attributes=dict(tnPkiKeyRingName=https.get("admin_key_ring")))))

            if https.get("client_certificate_trustpoint"):
                https_config["commHttps"]["children"].append(
                    dict(commRsClientCertCA=dict(attributes=dict(tDn="uni/userext/pkiext/tp-{0}".format(https.get("client_certificate_trustpoint")))))
                )

            child_configs.append(https_config)

        if telnet:
            child_configs.append(
                dict(
                    commTelnet=dict(
                        attributes=dict(
                            adminSt=telnet.get("admin_state"),
                            port=telnet.get("port"),
                        )
                    )
                )
            )

        if ssh_web:
            child_configs.append(
                dict(
                    commShellinabox=dict(
                        attributes=dict(
                            adminSt=ssh_web.get("admin_state"),
                        )
                    )
                )
            )

        aci.payload(
            aci_class=aci_class,
            class_config=dict(
                name=name,
                descr=description,
                nameAlias=name_alias,
            ),
            child_configs=child_configs,
        )

        aci.get_diff(aci_class=aci_class)

        aci.post_config()

    elif state == "absent":
        aci.delete_config()

    aci.exit_json()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

# Copyright: Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

#
# Note: This code should probably live in amazon.aws rather than community.aws.
# However, for the sake of getting something into a useful shape first, it makes
# sense for it to start life in community.aws.
#

# import typing
from copy import deepcopy

try:
    import botocore.waiter as botocore_waiter

    import_error = None
except ImportError as e:
    botocore_waiter = None
    import_error = e


class BaseWaiterFactory:
    """
    A helper class used for creating additional waiters.
    Unlike the waiters available directly from botocore these waiters will
    automatically retry on common (temporary) AWS failures.

    This class should be treated as an abstract class and subclassed before use.
    A subclass should:
    - override _BaseWaiterFactory._waiter_model_data to return the data defining
      the waiter

    Usage:
    waiter_factory = BaseWaiterFactory()
    waiter = waiter_factory.get_waiter(client, 'my_waiter_name')
    waiter.wait(**params)
    """

    def __init__(self):
        if not botocore_waiter:
            return

        # While it would be nice to supliment this with the upstream data,
        # unfortunately client doesn't have a public method for getting the
        # waiter configs.
        data = self._inject_ratelimit_retries(self._waiter_model_data)
        self._model = botocore_waiter.WaiterModel(
            waiter_config=dict(version=2, waiters=data),
        )

    @property
    def _waiter_model_data(self):
        r"""
        Subclasses should override this method to return a dictionary mapping
        waiter names to the waiter definition.

        This data is similar to the data found in botocore's waiters-2.json
        files (for example: botocore/botocore/data/ec2/2016-11-15/waiters-2.json)
        with two differences:
        1) Waiter names do not have transformations applied during lookup
        2) Only the 'waiters' data is required, the data is assumed to be
           version 2

        for example:

        @property
        def _waiter_model_data(self):
            return dict(
                tgw_attachment_deleted=dict(
                    operation='DescribeTransitGatewayAttachments',
                    delay=5, maxAttempts=120,
                    acceptors=[
                        dict(state='retry', matcher='pathAll', expected='deleting', argument='TransitGatewayAttachments[].State'),
                        dict(state='success', matcher='pathAll', expected='deleted', argument='TransitGatewayAttachments[].State'),
                        dict(state='success', matcher='path', expected=True, argument='length(TransitGatewayAttachments[]) == `0`'),
                        dict(state='success', matcher='error', expected='InvalidRouteTableID.NotFound'),
                    ]
                ),
            )

        or

        @property
        def _waiter_model_data(self):
            return {
                "instance_exists": {
                    "delay": 5,
                    "maxAttempts": 40,
                    "operation": "DescribeInstances",
                    "acceptors": [
                        {
                            "matcher": "path",
                            "expected": true,
                            "argument": "length(Reservations[]) > `0`",
                            "state": "success"
                        },
                        {
                            "matcher": "error",
                            "expected": "InvalidInstanceID.NotFound",
                            "state": "retry"
                        }
                    ]
                },
            }
        """

        return dict()

    def _inject_ratelimit_retries(self, model_data):
        extra_retries = [
            "RequestLimitExceeded",
            "Unavailable",
            "ServiceUnavailable",
            "InternalFailure",
            "InternalError",
            "TooManyRequestsException",
            "Throttling",
        ]

        acceptors = []
        for error in extra_retries:
            acceptors.append(dict(state="retry", matcher="error", expected=error))

        _model_data = deepcopy(model_data)
        for waiter_name in _model_data:
            _model_data[waiter_name]["acceptors"].extend(acceptors)

        return _model_data

    def get_waiter(self, client, waiter_name: str):
        if import_error:
            # We shouldn't get here, but if someone's trying to use this without botocore installed
            # let's re-raise the actual import error
            raise import_error

        waiters = self._model.waiter_names
        if waiter_name not in waiters:
            raise NotImplementedError(f"Unable to find waiter {waiter_name}.  Available_waiters: {waiters}")
        return botocore_waiter.create_waiter_with_client(
            waiter_name,
            self._model,
            client,
        )


def custom_waiter_config(timeout: int, default_pause: int = 2):
    """
    Generates the waiter_config dict that allows configuring a custom wait_timeout

    Where the pause and the timeouts aren't perfectly divisible, this will default to waiting
    slightly longer than the configured timeout so that we give at least the timeout time for a
    change to happen.
    """

    pause = default_pause

    # Do something sensible when the user's passed a short timeout, but our default pause wouldn't
    # have allowed any retries
    if timeout < (default_pause * 3):
        pause = max(1, timeout // 3)

    attempts = 1 + (timeout // pause)

    if (attempts - 1) * pause < timeout:
        attempts += 1

    return dict(Delay=pause, MaxAttempts=attempts)



#!/usr/bin/python

#  Copyright Edward Hilgendorf, <edward@hilgendorf.me>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: dnsimple_info

short_description: Pull basic info from DNSimple API

version_added: "4.2.0"

description: Retrieve existing records and domains from DNSimple API.

extends_documentation_fragment:
  - community.general._attributes
  - community.general._attributes.info_module

options:
  name:
    description:
      - The domain name to retrieve info from.
      - Returns all associated records for this domain if specified.
      - If not specified, returns all domains associated with the account ID.
    type: str

  account_id:
    description: The account ID to query.
    required: true
    type: str

  api_key:
    description: The API key to use.
    required: true
    type: str

  record:
    description:
      - The record to find.
      - If specified, only this record is returned instead of all records.
    type: str

  sandbox:
    description: Whether or not to use sandbox environment.
    default: false
    type: bool

author:
  - Edward Hilgendorf (@edhilgendorf)
"""

EXAMPLES = r"""
- name: Get all domains from an account
  community.general.dnsimple_info:
    account_id: "1234"
    api_key: "1234"

- name: Get all records from a domain
  community.general.dnsimple_info:
    name: "example.com"
    account_id: "1234"
    api_key: "1234"

- name: Get all info from a matching record
  community.general.dnsimple_info:
    name: "example.com"
    record: "subdomain"
    account_id: "1234"
    api_key: "1234"
"""

RETURN = r"""
dnsimple_domain_info:
  description: Returns a list of dictionaries of all domains associated with the supplied account ID.
  type: list
  elements: dict
  returned: success when O(name) is not specified
  sample:
    - account_id: 1234
      created_at: '2021-10-16T21:25:42Z'
      id: 123456
      last_transferred_at:
      name: example.com
      reverse: false
      secondary: false
      updated_at: '2021-11-10T20:22:50Z'
  contains:
    account_id:
      description: The account ID.
      type: int
    created_at:
      description: When the domain entry was created.
      type: str
    id:
      description: ID of the entry.
      type: int
    last_transferred_at:
      description: Date the domain was transferred, or empty if not.
      type: str
    name:
      description: Name of the record.
      type: str
    reverse:
      description: Whether or not it is a reverse zone record.
      type: bool
    updated_at:
      description: When the domain entry was updated.
      type: str

dnsimple_records_info:
  description: Returns a list of dictionaries with all records for the domain supplied.
  type: list
  elements: dict
  returned: success when O(name) is specified, but O(record) is not
  sample:
    - content: ns1.dnsimple.com admin.dnsimple.com
      created_at: '2021-10-16T19:07:34Z'
      id: 12345
      name: 'catheadbiscuit'
      parent_id:
      priority:
      regions:
        - global
      system_record: true
      ttl: 3600
      type: SOA
      updated_at: '2021-11-15T23:55:51Z'
      zone_id: example.com
  contains:
    content:
      description: Content of the returned record.
      type: str
    created_at:
      description: When the domain entry was created.
      type: str
    id:
      description: ID of the entry.
      type: int
    name:
      description: Name of the record.
      type: str
    parent_id:
      description: Parent record or null.
      type: int
    priority:
      description: Priority setting of the record.
      type: str
    regions:
      description: List of regions where the record is available.
      type: list
    system_record:
      description: Whether or not it is a system record.
      type: bool
    ttl:
      description: Record TTL.
      type: int
    type:
      description: Record type.
      type: str
    updated_at:
      description: When the domain entry was updated.
      type: str
    zone_id:
      description: ID of the zone that the record is associated with.
      type: str
dnsimple_record_info:
  description: Returns a list of dictionaries that match the record supplied.
  returned: success when O(name) and O(record) are specified
  type: list
  elements: dict
  sample:
    - content: 1.2.3.4
      created_at: '2021-11-15T23:55:51Z'
      id: 123456
      name: catheadbiscuit
      parent_id:
      priority:
      regions:
        - global
      system_record: false
      ttl: 3600
      type: A
      updated_at: '2021-11-15T23:55:51Z'
      zone_id: example.com
  contains:
    content:
      description: Content of the returned record.
      type: str
    created_at:
      description: When the domain entry was created.
      type: str
    id:
      description: ID of the entry.
      type: int
    name:
      description: Name of the record.
      type: str
    parent_id:
      description: Parent record or null.
      type: int
    priority:
      description: Priority setting of the record.
      type: str
    regions:
      description: List of regions where the record is available.
      type: list
    system_record:
      description: Whether or not it is a system record.
      type: bool
    ttl:
      description: Record TTL.
      type: int
    type:
      description: Record type.
      type: str
    updated_at:
      description: When the domain entry was updated.
      type: str
    zone_id:
      description: ID of the zone that the record is associated with.
      type: str
"""

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils import _deps as deps

with deps.declare("requests"):
    from requests import Request, Session


def build_url(account, key, is_sandbox):
    headers = {"Accept": "application/json", "Authorization": f"Bearer {key}"}
    sandbox = ".sandbox" if is_sandbox else ""
    url = f"https://api{sandbox}.dnsimple.com/v2/{account}"
    req = Request(url=url, headers=headers)
    prepped_request = req.prepare()
    return prepped_request


def iterate_data(module, request_object):
    base_url = request_object.url
    response = Session().send(request_object)
    if "pagination" not in response.json():
        module.fail_json("API Call failed, check ID, key and sandbox values")

    data = response.json()["data"]
    total_pages = response.json()["pagination"]["total_pages"]
    page = 1

    while page < total_pages:
        page = page + 1
        request_object.url = f"{base_url}&page={page}"
        new_results = Session().send(request_object)
        data = data + new_results.json()["data"]

    return data


def record_info(dnsimple_mod, req_obj):
    req_obj.url, req_obj.method = (
        f"{req_obj.url}/zones/{dnsimple_mod.params['name']}/records?name={dnsimple_mod.params['record']}",
        "GET",
    )
    return iterate_data(dnsimple_mod, req_obj)


def domain_info(dnsimple_mod, req_obj):
    req_obj.url, req_obj.method = f"{req_obj.url}/zones/{dnsimple_mod.params['name']}/records?per_page=100", "GET"
    return iterate_data(dnsimple_mod, req_obj)


def account_info(dnsimple_mod, req_obj):
    req_obj.url, req_obj.method = f"{req_obj.url}/zones/?per_page=100", "GET"
    return iterate_data(dnsimple_mod, req_obj)


def main():
    result = {"changed": False}

    module = AnsibleModule(
        argument_spec=dict(
            account_id=dict(required=True, type="str"),
            api_key=dict(required=True, type="str", no_log=True),
            name=dict(type="str"),
            record=dict(type="str"),
            sandbox=dict(type="bool", default=False),
        ),
        supports_check_mode=True,
    )

    params = module.params
    req = build_url(params["account_id"], params["api_key"], params["sandbox"])

    deps.validate(module)

    # If we have a record return info on that record
    if params["name"] and params["record"]:
        result["dnsimple_record_info"] = record_info(module, req)
        module.exit_json(**result)

        # If we have the account only and domain, return records for the domain
    elif params["name"]:
        result["dnsimple_records_info"] = domain_info(module, req)
        module.exit_json(**result)

        # If we have the account only, return domains
    else:
        result["dnsimple_domain_info"] = account_info(module, req)
        module.exit_json(**result)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}

DOCUMENTATION = r'''
---
module: intersight_ethernet_network_group_policy_info
short_description: Gather information about Ethernet Network Group Policies in Cisco Intersight
description:
  - Gather information about Ethernet Network Group Policies in L(Cisco Intersight,https://intersight.com).
  - Information can be filtered by O(organization) and O(name).
  - If no filters are passed, all Ethernet Network Group Policies will be returned.
extends_documentation_fragment: intersight
options:
  organization:
    description:
      - The name of the organization the Ethernet Network Group Policy belongs to.
    type: str
  name:
    description:
      - The name of the Ethernet Network Group Policy to gather information from.
    type: str
author:
  - Ron Gershburg (@rgershbu)
'''

EXAMPLES = r'''
- name: Fetch a specific Ethernet Network Group Policy by name
  cisco.intersight.intersight_ethernet_network_group_policy_info:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    name: "qinq-policy"

- name: Fetch all Ethernet Network Group Policies in a specific Organization
  cisco.intersight.intersight_ethernet_network_group_policy_info:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    organization: "default"

- name: Fetch all Ethernet Network Group Policies
  cisco.intersight.intersight_ethernet_network_group_policy_info:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
'''

RETURN = r'''
api_response:
  description: The API response output returned by the specified resource.
  returned: always
  type: dict
  sample:
    "api_response": [
    {
        "Name": "regular-vlans-policy",
        "ObjectType": "fabric.EthNetworkGroupPolicy",
        "VlanSettings": {
            "QinqEnabled": false,
            "AllowedVlans": "1-8,12,16"
        },
        "Tags": [
            {
                "Key": "Environment",
                "Value": "Production"
            }
        ]
    },
    {
        "Name": "qinq-policy",
        "ObjectType": "fabric.EthNetworkGroupPolicy",
        "VlanSettings": {
            "QinqEnabled": true,
            "QinqVlan": 4,
            "NativeVlan": 1
        },
        "Tags": [
            {
                "Key": "Site",
                "Value": "DataCenter-A"
            },
            {
                "Key": "Application",
                "Value": "Database"
            }
        ]
    }
  ]
'''


from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.intersight.plugins.module_utils.intersight import IntersightModule, intersight_argument_spec


def main():
    argument_spec = intersight_argument_spec.copy()
    argument_spec.update(
        organization=dict(type='str'),
        name=dict(type='str')
    )
    module = AnsibleModule(
        argument_spec,
        supports_check_mode=True,
    )

    intersight = IntersightModule(module)
    intersight.result['api_response'] = {}
    intersight.result['trace_id'] = ''

    resource_path = '/fabric/EthNetworkGroupPolicies'

    query_params = intersight.set_query_params()

    intersight.get_resource(
        resource_path=resource_path,
        query_params=query_params,
        return_list=True
    )

    module.exit_json(**intersight.result)


if __name__ == '__main__':
    main()



#!/usr/bin/python

# Copyright (c) 2021, Lennert Mertens (lennert@nubera.be)
# Copyright (c) 2019, Guillaume Martinez (lunik@tiwabbit.fr)
# Copyright (c) 2015, Werner Dijkerman (ikben@werner-dijkerman.nl)
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: gitlab_user
short_description: Creates/updates/deletes/blocks/unblocks GitLab Users
description:
  - When the user does not exist in GitLab, it is created.
  - When the user exists and state=absent, the user is deleted.
  - When the user exists and state=blocked, the user is blocked.
  - When changes are made to user, the user is updated.
notes:
  - From community.general 0.2.0 and onwards, name, email and password are optional while deleting the user.
author:
  - Werner Dijkerman (@dj-wasabi)
  - Guillaume Martinez (@Lunik)
  - Lennert Mertens (@LennertMertens)
  - Stef Graces (@stgrace)
requirements:
  - python-gitlab python module
  - administrator rights on the GitLab server
extends_documentation_fragment:
  - community.general._auth_basic
  - community.general._gitlab
  - community.general._attributes

attributes:
  check_mode:
    support: full
  diff_mode:
    support: none

options:
  name:
    description:
      - Name of the user you want to create.
      - Required only if O(state=present).
    type: str
  username:
    description:
      - The username of the user.
    required: true
    type: str
  password:
    description:
      - The password of the user.
      - GitLab server enforces minimum password length to 8, set this value with 8 or more characters.
    type: str
  reset_password:
    description:
      - Whether the user can change its password or not.
    default: false
    type: bool
    version_added: 3.3.0
  email:
    description:
      - The email that belongs to the user.
      - Required only if O(state=present).
    type: str
  sshkey_name:
    description:
      - The name of the SSH public key.
    type: str
  sshkey_file:
    description:
      - The SSH public key itself.
    type: str
  sshkey_expires_at:
    description:
      - The expiration date of the SSH public key in ISO 8601 format C(YYYY-MM-DDTHH:MM:SSZ).
      - This is only used when creating or replacing SSH public keys.
    type: str
    version_added: 3.1.0
  sshkey_update_mode:
    description:
      - Controls how existing SSH keys with the same O(sshkey_name) are handled.
      - Use V(create) to preserve the historical behavior and leave existing keys with the same name unchanged.
      - Use V(update) to replace an existing key only when exactly one key with the same name exists and its key material differs.
      - Use V(deduplicate) to delete all keys with the same name before creating the requested key.
      - SSH key comments are ignored when comparing key material.
    type: str
    choices: ["create", "update", "deduplicate"]
    default: create
    version_added: 13.0.0
  group:
    description:
      - ID or Full path of parent group in the form of group/name.
      - Add user as a member to this group.
    type: str
  access_level:
    description:
      - The access level to the group.
      - The value V(master) is an alias for V(maintainer).
    default: guest
    type: str
    choices: ["guest", "reporter", "developer", "master", "maintainer", "owner"]
  state:
    description:
      - Create, delete or block a user.
    default: present
    type: str
    choices: ["present", "absent", "blocked", "unblocked"]
  confirm:
    description:
      - Require confirmation.
    type: bool
    default: true
  isadmin:
    description:
      - Grant admin privileges to the user.
    type: bool
    default: false
  external:
    description:
      - Define external parameter for this user.
    type: bool
    default: false
  identities:
    description:
      - List of identities to be added/updated for this user.
      - To remove all other identities from this user, set O(overwrite_identities=true).
    type: list
    elements: dict
    suboptions:
      provider:
        description:
          - The name of the external identity provider.
        type: str
      extern_uid:
        description:
          - User ID for external identity.
        type: str
    version_added: 3.3.0
  overwrite_identities:
    description:
      - Overwrite identities with identities added in this module.
      - This means that all identities that the user has and that are not listed in O(identities) are removed from the user.
      - This is only done if a list is provided for O(identities). To remove all identities, provide an empty list.
    type: bool
    default: false
    version_added: 3.3.0
"""

EXAMPLES = r"""
- name: "Delete GitLab User"
  community.general.gitlab_user:
    api_url: https://gitlab.example.com/
    api_token: "{{ access_token }}"
    username: myusername
    state: absent

- name: "Create GitLab User"
  community.general.gitlab_user:
    api_url: https://gitlab.example.com/
    validate_certs: true
    api_username: dj-wasabi
    api_password: "MySecretPassword"
    name: My Name
    username: myusername
    password: mysecretpassword
    email: me@example.com
    sshkey_name: MySSH
    sshkey_file: ssh-rsa AAAAB3NzaC1yc...
    state: present
    group: super_group/mon_group
    access_level: owner

- name: "Replace a GitLab SSH key when one matching name exists"
  community.general.gitlab_user:
    api_url: https://gitlab.example.com/
    api_token: "{{ access_token }}"
    name: My Name
    username: myusername
    email: me@example.com
    sshkey_name: MySSH
    sshkey_file: ssh-ed25519 AAAAC3NzaC1lZDI1NTE5AAAAI...
    sshkey_update_mode: update
    state: present

- name: "Create GitLab User using external identity provider"
  community.general.gitlab_user:
    api_url: https://gitlab.example.com/
    validate_certs: true
    api_token: "{{ access_token }}"
    name: My Name
    username: myusername
    password: mysecretpassword
    email: me@example.com
    identities:
      - provider: Keycloak
        extern_uid: f278f95c-12c7-4d51-996f-758cc2eb11bc
    state: present
    group: super_group/mon_group
    access_level: owner

- name: "Block GitLab User"
  community.general.gitlab_user:
    api_url: https://gitlab.example.com/
    api_token: "{{ access_token }}"
    username: myusername
    state: blocked

- name: "Unblock GitLab User"
  community.general.gitlab_user:
    api_url: https://gitlab.example.com/
    api_token: "{{ access_token }}"
    username: myusername
    state: unblocked
"""

RETURN = r"""
msg:
  description: Success or failure message.
  returned: always
  type: str
  sample: "Success"

result:
  description: JSON-parsed response from the server.
  returned: always
  type: dict

error:
  description: The error message returned by the GitLab API.
  returned: failed
  type: str
  sample: "400: path is already in use"

user:
  description: API object.
  returned: always
  type: dict
"""


from ansible.module_utils.api import basic_auth_argument_spec
from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._gitlab import (
    auth_argument_spec,
    find_group,
    gitlab,
    gitlab_authentication,
    list_all_kwargs,
)


class GitLabUser:
    def __init__(self, module, gitlab_instance):
        self._module = module
        self._gitlab = gitlab_instance
        self.user_object = None
        self.ACCESS_LEVEL = {
            "guest": gitlab.const.GUEST_ACCESS,
            "reporter": gitlab.const.REPORTER_ACCESS,
            "developer": gitlab.const.DEVELOPER_ACCESS,
            "master": gitlab.const.MAINTAINER_ACCESS,
            "maintainer": gitlab.const.MAINTAINER_ACCESS,
            "owner": gitlab.const.OWNER_ACCESS,
        }

    """
    @param username Username of the user
    @param options User options
    """

    def create_or_update_user(self, username, options):
        changed = False
        potentionally_changed = False

        # Because we have already call userExists in main()
        if self.user_object is None:
            user = self.create_user(
                {
                    "name": options["name"],
                    "username": username,
                    "password": options["password"],
                    "reset_password": options["reset_password"],
                    "email": options["email"],
                    "skip_confirmation": not options["confirm"],
                    "admin": options["isadmin"],
                    "external": options["external"],
                    "identities": options["identities"],
                }
            )
            changed = True
        else:
            changed, user = self.update_user(
                self.user_object,
                {
                    # add "normal" parameters here, put uncheckable
                    # params in the dict below
                    "name": {"value": options["name"]},
                    "email": {"value": options["email"]},
                    # note: for some attributes like this one the key
                    #   from reading back from server is unfortunately
                    #   different to the one needed for pushing/writing,
                    #   in that case use the optional setter key
                    "is_admin": {"value": options["isadmin"], "setter": "admin"},
                    "external": {"value": options["external"]},
                    "identities": {"value": options["identities"]},
                },
                {
                    # put "uncheckable" params here, this means params
                    # which the gitlab does accept for setting but does
                    # not return any information about it
                    "skip_reconfirmation": {"value": not options["confirm"]},
                    "password": {"value": options["password"]},
                    "reset_password": {"value": options["reset_password"]},
                    "overwrite_identities": {"value": options["overwrite_identities"]},
                },
            )

            # note: as we unfortunately have some uncheckable parameters
            #   where it is not possible to determine if the update
            #   changed something or not, we must assume here that a
            #   changed happened and that an user object update is needed
            potentionally_changed = True

        # Assign ssh keys
        if options["sshkey_name"] and options["sshkey_file"]:
            key_changed = self.add_ssh_key_to_user(
                user,
                {
                    "name": options["sshkey_name"],
                    "file": options["sshkey_file"],
                    "expires_at": options["sshkey_expires_at"],
                    "update_mode": options["sshkey_update_mode"],
                },
            )
            changed = changed or key_changed

        # Assign group
        if options["group_path"]:
            group_changed = self.assign_user_to_group(user, options["group_path"], options["access_level"])
            changed = changed or group_changed

        self.user_object = user
        if (changed or potentionally_changed) and not self._module.check_mode:
            try:
                user.save()
            except Exception as e:
                self._module.fail_json(msg=f"Failed to update user: {e} ")

        if changed:
            if self._module.check_mode:
                self._module.exit_json(changed=True, msg=f"Successfully created or updated the user {username}")
            return True
        else:
            return False

    """
    @param group User object
    """

    def get_user_id(self, user):
        if user is not None:
            return user.id
        return None

    """
    @param user User object
    @param sshkey_name Name of the ssh key
    """

    @staticmethod
    def normalize_ssh_key(ssh_key):
        return " ".join(ssh_key.split(None, 2)[:2])

    def find_ssh_keys(self, user, sshkey_name):
        return [key for key in user.keys.list(**list_all_kwargs) if key.title == sshkey_name]

    def ssh_key_exists(self, user, sshkey_name):
        return bool(self.find_ssh_keys(user, sshkey_name))

    """
    @param user User object
    @param sshkey Dict containing sshkey infos {"name": "", "file": "", "expires_at": ""}
    """

    def add_ssh_key_to_user(self, user, sshkey):
        user_keys = self.find_ssh_keys(user, sshkey["name"])
        desired_key = self.normalize_ssh_key(sshkey["file"])
        update_mode = sshkey["update_mode"]
        matching_keys = [key for key in user_keys if self.normalize_ssh_key(key.key) == desired_key]

        if update_mode == "create":
            if user_keys:
                return False
        elif update_mode == "update":
            if len(user_keys) > 1:
                self._module.warn(
                    f"Found multiple SSH keys named '{sshkey['name']}'. Skipping update because sshkey_update_mode=update requires a single matching key."
                )
                return False
            if matching_keys:
                return False
        elif len(user_keys) == 1 and matching_keys:
            return False

        if self._module.check_mode:
            return True

        try:
            for key in user_keys:
                key.delete()

            parameter = {
                "title": sshkey["name"],
                "key": sshkey["file"],
            }
            if sshkey["expires_at"] is not None:
                parameter["expires_at"] = sshkey["expires_at"]
            user.keys.create(parameter)
        except gitlab.exceptions.GitlabDeleteError as e:
            self._module.fail_json(msg=f"Failed to update sshkey for user: {e}")
        except gitlab.exceptions.GitlabCreateError as e:
            self._module.fail_json(msg=f"Failed to assign sshkey to user: {e}")
        return True

    """
    @param group Group object
    @param user_id Id of the user to find
    """

    def find_member(self, group, user_id):
        try:
            member = group.members.get(user_id)
        except gitlab.exceptions.GitlabGetError:
            return None
        return member

    """
    @param group Group object
    @param user_id Id of the user to check
    """

    def member_exists(self, group, user_id):
        member = self.find_member(group, user_id)

        return member is not None

    """
    @param group Group object
    @param user_id Id of the user to check
    @param access_level GitLab access_level to check
    """

    def member_as_good_access_level(self, group, user_id, access_level):
        member = self.find_member(group, user_id)

        return member.access_level == access_level

    """
    @param user User object
    @param group_path Complete path of the Group including parent group path. <parent_path>/<group_path>
    @param access_level GitLab access_level to assign
    """

    def assign_user_to_group(self, user, group_identifier, access_level):
        group = find_group(self._gitlab, group_identifier)

        if self._module.check_mode:
            return True

        if group is None:
            return False

        if self.member_exists(group, self.get_user_id(user)):
            member = self.find_member(group, self.get_user_id(user))
            if not self.member_as_good_access_level(group, member.id, self.ACCESS_LEVEL[access_level]):
                member.access_level = self.ACCESS_LEVEL[access_level]
                member.save()
                return True
        else:
            try:
                group.members.create(
                    {"user_id": self.get_user_id(user), "access_level": self.ACCESS_LEVEL[access_level]}
                )
            except gitlab.exceptions.GitlabCreateError as e:
                self._module.fail_json(msg=f"Failed to assign user to group: {e}")
            return True
        return False

    """
    @param user User object
    @param arguments User attributes
    """

    def update_user(self, user, arguments, uncheckable_args):
        changed = False

        for arg_key, arg_value in arguments.items():
            av = arg_value["value"]

            if av is not None:
                if arg_key == "identities":
                    changed = self.add_identities(user, av, uncheckable_args["overwrite_identities"]["value"])

                elif getattr(user, arg_key) != av:
                    setattr(user, arg_value.get("setter", arg_key), av)
                    changed = True

        for arg_key, arg_value in uncheckable_args.items():
            av = arg_value["value"]

            if av is not None:
                setattr(user, arg_value.get("setter", arg_key), av)

        return (changed, user)

    """
    @param arguments User attributes
    """

    def create_user(self, arguments):
        if self._module.check_mode:
            return True

        identities = None
        if "identities" in arguments:
            identities = arguments["identities"]
            del arguments["identities"]

        try:
            user = self._gitlab.users.create(arguments)
            if identities:
                self.add_identities(user, identities)

        except gitlab.exceptions.GitlabCreateError as e:
            self._module.fail_json(msg=f"Failed to create user: {e}")

        return user

    """
    @param user User object
    @param identities List of identities to be added/updated
    @param overwrite_identities Overwrite user identities with identities passed to this module
    """

    def add_identities(self, user, identities, overwrite_identities=False):
        changed = False
        if overwrite_identities:
            changed = self.delete_identities(user, identities)

        for identity in identities:
            if identity not in user.identities:
                user.provider = identity["provider"]
                user.extern_uid = identity["extern_uid"]
                if not self._module.check_mode:
                    user.save()
                changed = True
        return changed

    """
    @param user User object
    @param identities List of identities to be added/updated
    """

    def delete_identities(self, user, identities):
        changed = False
        for identity in user.identities:
            if identity not in identities:
                if not self._module.check_mode:
                    user.identityproviders.delete(identity["provider"])
                changed = True
        return changed

    """
    @param username Username of the user
    """

    def find_user(self, username):
        return next(
            (user for user in self._gitlab.users.list(search=username, **list_all_kwargs) if user.username == username),
            None,
        )

    """
    @param username Username of the user
    """

    def exists_user(self, username):
        # When user exists, object will be stored in self.user_object.
        user = self.find_user(username)
        if user:
            self.user_object = user
            return True
        return False

    """
    @param username Username of the user
    """

    def is_active(self, username):
        user = self.find_user(username)
        return user.attributes["state"] == "active"

    def delete_user(self):
        if self._module.check_mode:
            return True

        user = self.user_object

        return user.delete()

    def block_user(self):
        if self._module.check_mode:
            return True

        user = self.user_object

        return user.block()

    def unblock_user(self):
        if self._module.check_mode:
            return True

        user = self.user_object

        return user.unblock()


def sanitize_arguments(arguments):
    for key, value in list(arguments.items()):
        if value is None:
            del arguments[key]
    return arguments


def main():
    argument_spec = basic_auth_argument_spec()
    argument_spec.update(auth_argument_spec())
    argument_spec.update(
        dict(
            name=dict(type="str"),
            state=dict(type="str", default="present", choices=["absent", "present", "blocked", "unblocked"]),
            username=dict(type="str", required=True),
            password=dict(type="str", no_log=True),
            reset_password=dict(type="bool", default=False, no_log=False),
            email=dict(type="str"),
            sshkey_name=dict(type="str"),
            sshkey_file=dict(type="str", no_log=False),
            sshkey_expires_at=dict(type="str", no_log=False),
            sshkey_update_mode=dict(type="str", default="create", choices=["create", "update", "deduplicate"]),
            group=dict(type="str"),
            access_level=dict(
                type="str", default="guest", choices=["developer", "guest", "maintainer", "master", "owner", "reporter"]
            ),
            confirm=dict(type="bool", default=True),
            isadmin=dict(type="bool", default=False),
            external=dict(type="bool", default=False),
            identities=dict(type="list", elements="dict"),
            overwrite_identities=dict(type="bool", default=False),
        )
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        mutually_exclusive=[
            ["api_username", "api_token"],
            ["api_username", "api_oauth_token"],
            ["api_username", "api_job_token"],
            ["api_token", "api_oauth_token"],
            ["api_token", "api_job_token"],
        ],
        required_together=[
            ["api_username", "api_password"],
        ],
        required_one_of=[["api_username", "api_token", "api_oauth_token", "api_job_token"]],
        supports_check_mode=True,
        required_if=(("state", "present", ["name", "email"]),),
    )

    # check prerequisites and connect to gitlab server
    gitlab_instance = gitlab_authentication(module)

    user_name = module.params["name"]
    state = module.params["state"]
    user_username = module.params["username"].lower()
    user_password = module.params["password"]
    user_reset_password = module.params["reset_password"]
    user_email = module.params["email"]
    user_sshkey_name = module.params["sshkey_name"]
    user_sshkey_file = module.params["sshkey_file"]
    user_sshkey_expires_at = module.params["sshkey_expires_at"]
    user_sshkey_update_mode = module.params["sshkey_update_mode"]
    group_path = module.params["group"]
    access_level = module.params["access_level"]
    confirm = module.params["confirm"]
    user_isadmin = module.params["isadmin"]
    user_external = module.params["external"]
    user_identities = module.params["identities"]
    overwrite_identities = module.params["overwrite_identities"]

    gitlab_user = GitLabUser(module, gitlab_instance)
    user_exists = gitlab_user.exists_user(user_username)
    if user_exists:
        user_is_active = gitlab_user.is_active(user_username)
    else:
        user_is_active = False

    if state == "absent":
        if user_exists:
            gitlab_user.delete_user()
            module.exit_json(changed=True, msg=f"Successfully deleted user {user_username}")
        else:
            module.exit_json(changed=False, msg="User deleted or does not exists")

    if state == "blocked":
        if user_exists and user_is_active:
            gitlab_user.block_user()
            module.exit_json(changed=True, msg=f"Successfully blocked user {user_username}")
        else:
            module.exit_json(changed=False, msg="User already blocked or does not exists")

    if state == "unblocked":
        if user_exists and not user_is_active:
            gitlab_user.unblock_user()
            module.exit_json(changed=True, msg=f"Successfully unblocked user {user_username}")
        else:
            module.exit_json(changed=False, msg="User is not blocked or does not exists")

    if state == "present":
        if gitlab_user.create_or_update_user(
            user_username,
            {
                "name": user_name,
                "password": user_password,
                "reset_password": user_reset_password,
                "email": user_email,
                "sshkey_name": user_sshkey_name,
                "sshkey_file": user_sshkey_file,
                "sshkey_expires_at": user_sshkey_expires_at,
                "sshkey_update_mode": user_sshkey_update_mode,
                "group_path": group_path,
                "access_level": access_level,
                "confirm": confirm,
                "isadmin": user_isadmin,
                "external": user_external,
                "identities": user_identities,
                "overwrite_identities": overwrite_identities,
            },
        ):
            module.exit_json(
                changed=True,
                msg=f"Successfully created or updated the user {user_username}",
                user=gitlab_user.user_object._attrs,
            )
        else:
            module.exit_json(
                changed=False, msg=f"No need to update the user {user_username}", user=gitlab_user.user_object._attrs
            )


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright (c) Benjamin Jolivot <bjolivot@gmail.com>
# Inspired by slack module :
#    # Copyright (c) 2017, Steve Pletcher <steve@steve-pletcher.com>
#    # Copyright (c) 2016, René Moser <mail@renemoser.net>
#    # Copyright (c) 2015, Stefan Berggren <nsg@nsg.cc>
#    # Copyright (c) 2014, Ramon de la Fuente <ramon@delafuente.nl>)
#
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: mattermost
short_description: Send Mattermost notifications
description:
  - Sends notifications to U(http://your.mattermost.url) using the Incoming WebHook integration.
author: "Benjamin Jolivot (@bjolivot)"
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  url:
    type: str
    description:
      - Mattermost URL (for example V(http://mattermost.yourcompany.com)).
    required: true
  api_key:
    type: str
    description:
      - Mattermost webhook API key. Log into your Mattermost site, go to Menu -> Integration -> Incoming Webhook -> Add Incoming
        Webhook. This gives you a full URL. O(api_key) is the last part. U(http://mattermost.example.com/hooks/API_KEY).
    required: true
  text:
    type: str
    description:
      - Text to send. Note that the module does not handle escaping characters.
      - Required when O(attachments) is not set.
  attachments:
    type: list
    elements: dict
    description:
      - Define a list of attachments.
      - For more information, see U(https://developers.mattermost.com/integrate/admin-guide/admin-message-attachments/).
      - Required when O(text) is not set.
    version_added: 4.3.0
  channel:
    type: str
    description:
      - Channel to send the message to. If absent, the message goes to the channel selected for the O(api_key).
  username:
    type: str
    description:
      - This is the sender of the message (Username Override need to be enabled by mattermost admin, see mattermost doc).
    default: Ansible
  icon_url:
    type: str
    description:
      - URL for the message sender's icon.
    default: https://docs.ansible.com/favicon/favicon.ico
  priority:
    type: str
    description:
      - Set a priority for the message.
    choices: [important, urgent]
    version_added: 10.0.0
  validate_certs:
    description:
      - If V(false), SSL certificates are not validated. This should only be used on personally controlled sites using self-signed
        certificates.
    default: true
    type: bool
"""

EXAMPLES = r"""
- name: Send notification message via Mattermost
  community.general.mattermost:
    url: http://mattermost.example.com
    api_key: my_api_key
    text: '{{ inventory_hostname }} completed'

- name: Send notification message via Mattermost all options
  community.general.mattermost:
    url: http://mattermost.example.com
    api_key: my_api_key
    text: '{{ inventory_hostname }} completed'
    channel: notifications
    username: 'Ansible on {{ inventory_hostname }}'
    icon_url: http://www.example.com/some-image-file.png
    priority: important

- name: Send attachments message via Mattermost
  community.general.mattermost:
    url: http://mattermost.example.com
    api_key: my_api_key
    attachments:
      - text: Display my system load on host A and B
        color: '#ff00dd'
        title: System load
        fields:
          - title: System A
            value: "load average: 0,74, 0,66, 0,63"
            short: true
          - title: System B
            value: 'load average: 5,16, 4,64, 2,43'
            short: true
"""

RETURN = r"""
payload:
  description: Mattermost payload.
  returned: success
  type: str
webhook_url:
  description: URL the webhook is sent to.
  returned: success
  type: str
"""

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.urls import fetch_url


def main():
    module = AnsibleModule(
        supports_check_mode=True,
        argument_spec=dict(
            url=dict(type="str", required=True),
            api_key=dict(type="str", required=True, no_log=True),
            text=dict(type="str"),
            channel=dict(type="str"),
            username=dict(type="str", default="Ansible"),
            icon_url=dict(type="str", default="https://docs.ansible.com/favicon/favicon.ico"),
            priority=dict(type="str", choices=["important", "urgent"]),
            validate_certs=dict(default=True, type="bool"),
            attachments=dict(type="list", elements="dict"),
        ),
        required_one_of=[
            ("text", "attachments"),
        ],
    )
    # init return dict
    result = dict(changed=False, msg="OK")

    # define webhook
    webhook_url = f"{module.params['url']}/hooks/{module.params['api_key']}"
    result["webhook_url"] = webhook_url

    # define payload
    payload = {}
    for param in ["text", "channel", "username", "icon_url", "attachments"]:
        if module.params[param] is not None:
            payload[param] = module.params[param]
    if module.params["priority"] is not None:
        payload["priority"] = {"priority": module.params["priority"]}

    payload = module.jsonify(payload)
    result["payload"] = payload

    # http headers
    headers = {
        "Content-Type": "application/json",
        "Accept": "application/json",
    }

    # notes:
    # Nothing is done in check mode
    # it'll pass even if your server is down or/and if your token is invalid.
    # If someone find good way to check...

    # send request if not in test mode
    if module.check_mode is False:
        response, info = fetch_url(module=module, url=webhook_url, headers=headers, method="POST", data=payload)

        # something's wrong
        if info["status"] != 200:
            # some problem
            result["msg"] = f"Failed to send mattermost message, the error was: {info['msg']}"
            module.fail_json(**result)

    # Looks good
    module.exit_json(**result)


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright (c) 2015, Tim Hoiberg <tim.hoiberg@gmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: bundler
short_description: Manage Ruby Gem dependencies with Bundler
description:
  - Manage installation and Gem version dependencies for Ruby using the Bundler gem.
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  executable:
    type: str
    description:
      - The path to the bundler executable.
  state:
    type: str
    description:
      - The desired state of the Gem bundle. V(latest) updates gems to the most recent, acceptable version.
    choices: [present, latest]
    default: present
  chdir:
    type: path
    description:
      - The directory to execute the bundler commands from. This directory needs to contain a valid Gemfile or .bundle/ directory.
      - If not specified, it defaults to the temporary working directory.
  exclude_groups:
    type: list
    elements: str
    description:
      - A list of Gemfile groups to exclude during operations. This only applies when O(state=present). Bundler considers
        this a 'remembered' property for the Gemfile and automatically excludes groups in future operations even if O(exclude_groups)
        is not set.
  clean:
    description:
      - Only applies if O(state=present). If set removes any gems on the target host that are not in the gemfile.
    type: bool
    default: false
  gemfile:
    type: path
    description:
      - Only applies if O(state=present). The path to the gemfile to use to install gems.
      - If not specified it defaults to the Gemfile in current directory.
  local:
    description:
      - If set only installs gems from the cache on the target host.
    type: bool
    default: false
  deployment_mode:
    description:
      - Only applies if O(state=present). If set it installs gems in C(./vendor/bundle) instead of the default location. Requires
        a C(Gemfile.lock) file to have been created prior.
    type: bool
    default: false
  user_install:
    description:
      - Only applies if O(state=present). Installs gems in the local user's cache or for all users.
    type: bool
    default: true
  gem_path:
    type: path
    description:
      - Only applies if O(state=present). Specifies the directory to install the gems into. If O(chdir) is set then this path
        is relative to O(chdir).
      - If not specified the default RubyGems gem paths are used.
  binstub_directory:
    type: path
    description:
      - Only applies if O(state=present). Specifies the directory to install any gem bins files to. When executed the bin
        files run within the context of the Gemfile and fail if any required gem dependencies are not installed. If O(chdir)
        is set then this path is relative to O(chdir).
  extra_args:
    type: str
    description:
      - A space separated string of additional commands that can be applied to the Bundler command. Refer to the Bundler documentation
        for more information.
author: "Tim Hoiberg (@thoiberg)"
"""

EXAMPLES = r"""
- name: Install gems from a Gemfile in the current directory
  community.general.bundler:
    state: present
    executable: ~/.rvm/gems/2.1.5/bin/bundle

- name: Exclude the production group from installing
  community.general.bundler:
    state: present
    exclude_groups: production

- name: Install gems into ./vendor/bundle
  community.general.bundler:
    state: present
    deployment_mode: true

- name: Install gems using a Gemfile in another directory
  community.general.bundler:
    state: present
    gemfile: ../rails_project/Gemfile

- name: Update Gemfile in another directory
  community.general.bundler:
    state: latest
    chdir: ~/rails_project
"""

from ansible.module_utils.basic import AnsibleModule


def get_bundler_executable(module):
    if module.params.get("executable"):
        result = module.params.get("executable").split(" ")
    else:
        result = [module.get_bin_path("bundle", True)]
    return result


def main():
    module = AnsibleModule(
        argument_spec=dict(
            executable=dict(),
            state=dict(default="present", choices=["present", "latest"]),
            chdir=dict(type="path"),
            exclude_groups=dict(type="list", elements="str"),
            clean=dict(default=False, type="bool"),
            gemfile=dict(type="path"),
            local=dict(default=False, type="bool"),
            deployment_mode=dict(default=False, type="bool"),
            user_install=dict(default=True, type="bool"),
            gem_path=dict(type="path"),
            binstub_directory=dict(type="path"),
            extra_args=dict(),
        ),
        supports_check_mode=True,
    )
    module.run_command_environ_update = {"LANGUAGE": "C", "LC_ALL": "C"}

    state = module.params.get("state")
    chdir = module.params.get("chdir")
    exclude_groups = module.params.get("exclude_groups")
    clean = module.params.get("clean")
    gemfile = module.params.get("gemfile")
    local = module.params.get("local")
    deployment_mode = module.params.get("deployment_mode")
    user_install = module.params.get("user_install")
    gem_path = module.params.get("gem_path")
    binstub_directory = module.params.get("binstub_directory")
    extra_args = module.params.get("extra_args")

    cmd = get_bundler_executable(module)

    if module.check_mode:
        cmd.append("check")
        rc, out, err = module.run_command(cmd, cwd=chdir, check_rc=False)

        module.exit_json(changed=rc != 0, state=state, stdout=out, stderr=err)

    bundle_env = {}
    if state == "present":
        cmd.append("install")
        if exclude_groups:
            bundle_env["BUNDLE_WITHOUT"] = ":".join(exclude_groups)
        if clean:
            bundle_env["BUNDLE_CLEAN"] = "true"
        if gemfile:
            cmd.extend(["--gemfile", gemfile])
        if local:
            cmd.append("--local")
        if deployment_mode:
            bundle_env["BUNDLE_DEPLOYMENT"] = "true"
        if not user_install:
            cmd.append("--system")
        if gem_path:
            bundle_env["BUNDLE_PATH"] = gem_path
        if binstub_directory:
            bundle_env["BUNDLE_BIN"] = binstub_directory
    else:
        cmd.append("update")
        if local:
            cmd.append("--local")

    if extra_args:
        cmd.extend(extra_args.split(" "))

    rc, out, err = module.run_command(cmd, cwd=chdir, check_rc=True, environ_update=bundle_env)

    module.exit_json(changed="Installing" in out, state=state, stdout=out, stderr=err)


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright (c) 2013, Jeroen Hoekx <jeroen.hoekx@dsquare.be>, Alexander Bulimov <lazywolf0@gmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
author:
  - Jeroen Hoekx (@jhoekx)
  - Alexander Bulimov (@abulimov)
  - Raoul Baudach (@unkaputtbar112)
  - Ziga Kern (@zigaSRC)
module: lvol
short_description: Configure LVM logical volumes
description:
  - This module creates, removes or resizes logical volumes.
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  vg:
    type: str
    required: true
    description:
      - The volume group this logical volume is part of.
  lv:
    type: str
    description:
      - The name of the logical volume.
  size:
    type: str
    description:
      - The size of the logical volume, according to lvcreate(8) C(--size), by default in megabytes or optionally with one
        of [bBsSkKmMgGtTpPeE] units; or according to lvcreate(8) C(--extents) as a percentage of [VG|PVS|FREE|ORIGIN]; Float
        values must begin with a digit.
      - When resizing, apart from specifying an absolute size you may, according to lvextend(8)|lvreduce(8) C(--size), specify
        the amount to extend the logical volume with the prefix V(+) or the amount to reduce the logical volume by with prefix
        V(-).
      - Resizing using V(+) or V(-) was not supported prior to community.general 3.0.0.
      - Please note that when using V(+), V(-), or percentage of FREE, the module is B(not idempotent).
  state:
    type: str
    description:
      - Control if the logical volume exists. If V(present) and the volume does not already exist then the O(size) option
        is required.
    choices: [absent, present]
    default: present
  active:
    description:
      - Whether the volume is active and visible to the host.
    type: bool
    default: true
  force:
    description:
      - Shrink or remove operations of volumes requires this switch. Ensures that filesystems never get corrupted/destroyed
        by mistake.
    type: bool
    default: false
  opts:
    type: str
    description:
      - Free-form options to be passed to the lvcreate command.
  snapshot:
    type: str
    description:
      - The name of a snapshot volume to be configured. When creating a snapshot volume, the O(lv) parameter specifies the
        origin volume.
  pvs:
    type: list
    elements: str
    description:
      - List of physical volumes (for example V(/dev/sda, /dev/sdb)).
  thinpool:
    type: str
    description:
      - The thin pool volume name. When you want to create a thin provisioned volume, specify a thin pool volume name.
  shrink:
    description:
      - Shrink if current size is higher than size requested.
    type: bool
    default: true
  resizefs:
    description:
      - Resize the underlying filesystem together with the logical volume.
      - Supported for C(ext2), C(ext3), C(ext4), C(reiserfs) and C(XFS) filesystems. Attempts to resize other filesystem types
        result in failure.
    type: bool
    default: false
notes:
  - You must specify lv (when managing the state of logical volumes) or thinpool (when managing a thin provisioned volume).
"""

EXAMPLES = r"""
- name: Create a logical volume of 512m
  community.general.lvol:
    vg: firefly
    lv: test
    size: 512

- name: Create a logical volume of 512m with disks /dev/sda and /dev/sdb
  community.general.lvol:
    vg: firefly
    lv: test
    size: 512
    pvs:
      - /dev/sda
      - /dev/sdb

- name: Create cache pool logical volume
  community.general.lvol:
    vg: firefly
    lv: lvcache
    size: 512m
    opts: --type cache-pool

- name: Create a logical volume of 512g.
  community.general.lvol:
    vg: firefly
    lv: test
    size: 512g

- name: Create a logical volume the size of all remaining space in the volume group
  community.general.lvol:
    vg: firefly
    lv: test
    size: 100%FREE

- name: Create a logical volume with special options
  community.general.lvol:
    vg: firefly
    lv: test
    size: 512g
    opts: -r 16

- name: Extend the logical volume to 1024m.
  community.general.lvol:
    vg: firefly
    lv: test
    size: 1024

- name: Extend the logical volume to consume all remaining space in the volume group
  community.general.lvol:
    vg: firefly
    lv: test
    size: +100%FREE

- name: Extend the logical volume by given space
  community.general.lvol:
    vg: firefly
    lv: test
    size: +512M

- name: Extend the logical volume to take all remaining space of the PVs and resize the underlying filesystem
  community.general.lvol:
    vg: firefly
    lv: test
    size: 100%PVS
    resizefs: true

- name: Resize the logical volume to % of VG
  community.general.lvol:
    vg: firefly
    lv: test
    size: 80%VG
    force: true

- name: Reduce the logical volume to 512m
  community.general.lvol:
    vg: firefly
    lv: test
    size: 512
    force: true

- name: Reduce the logical volume by given space
  community.general.lvol:
    vg: firefly
    lv: test
    size: -512M
    force: true

- name: Set the logical volume to 512m and do not try to shrink if size is lower than current one
  community.general.lvol:
    vg: firefly
    lv: test
    size: 512
    shrink: false

- name: Remove the logical volume.
  community.general.lvol:
    vg: firefly
    lv: test
    state: absent
    force: true

- name: Create a snapshot volume of the test logical volume.
  community.general.lvol:
    vg: firefly
    lv: test
    snapshot: snap1
    size: 100m

- name: Deactivate a logical volume
  community.general.lvol:
    vg: firefly
    lv: test
    active: false

- name: Create a deactivated logical volume
  community.general.lvol:
    vg: firefly
    lv: test
    size: 512g
    active: false

- name: Create a thin pool of 512g
  community.general.lvol:
    vg: firefly
    thinpool: testpool
    size: 512g

- name: Create a thin volume of 128g
  community.general.lvol:
    vg: firefly
    lv: test
    thinpool: testpool
    size: 128g
"""

import re
import shlex

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._lvm import (
    lvchange_runner,
    lvcreate_runner,
    lvextend_runner,
    lvreduce_runner,
    lvremove_runner,
    lvs_runner,
    vgs_runner,
)


def mkversion(major, minor, patch):
    return (1000 * 1000 * int(major)) + (1000 * int(minor)) + int(patch)


def parse_lvs(data):
    lvs = []
    for line in data.splitlines():
        parts = line.strip().split(";")
        lvs.append(
            {
                "name": parts[0].replace("[", "").replace("]", ""),
                "size": float(parts[1]),
                "active": (parts[2][4] == "a"),
                "thinpool": (parts[2][0] == "t"),
                "thinvol": (parts[2][0] == "V"),
            }
        )
    return lvs


def parse_vgs(data):
    vgs = []
    for line in data.splitlines():
        parts = line.strip().split(";")
        vgs.append({"name": parts[0], "size": float(parts[1]), "free": float(parts[2]), "ext_size": float(parts[3])})
    return vgs


def get_lvm_version(module):
    ver_cmd = module.get_bin_path("lvm", required=True)
    rc, out, err = module.run_command([ver_cmd, "version"])
    if rc != 0:
        return None
    m = re.search(r"LVM version:\s+(\d+)\.(\d+)\.(\d+)", out)
    if not m:
        return None
    return mkversion(m.group(1), m.group(2), m.group(3))


def _run_resize(module, is_extend, lvextend, lvreduce, size_opt, size_value, resizefs, lv_name, pvs):
    size_key = f"size_{size_opt}"
    if is_extend:
        with lvextend(f"test resizefs {size_key} lv pvs") as ctx:
            return ctx.run(test=module.check_mode, resizefs=resizefs, **{size_key: size_value}, lv=[lv_name], pvs=pvs)
    else:
        with lvreduce(f"test force resizefs {size_key} lv pvs") as ctx:
            return ctx.run(
                test=module.check_mode, force=True, resizefs=resizefs, **{size_key: size_value}, lv=[lv_name], pvs=pvs
            )


def main():
    module = AnsibleModule(
        argument_spec=dict(
            vg=dict(type="str", required=True),
            lv=dict(type="str"),
            size=dict(type="str"),
            opts=dict(type="str"),
            state=dict(type="str", default="present", choices=["absent", "present"]),
            force=dict(type="bool", default=False),
            shrink=dict(type="bool", default=True),
            active=dict(type="bool", default=True),
            snapshot=dict(type="str"),
            pvs=dict(type="list", elements="str"),
            resizefs=dict(type="bool", default=False),
            thinpool=dict(type="str"),
        ),
        supports_check_mode=True,
        required_one_of=(["lv", "thinpool"],),
    )

    # Determine if the "--yes" option should be used
    version_found = get_lvm_version(module)
    if version_found is None:
        module.fail_json(msg="Failed to get LVM version number")
    version_yesopt = mkversion(2, 2, 99)  # First LVM with the "--yes" option
    use_yes = version_found >= version_yesopt

    vg = module.params["vg"]
    lv = module.params["lv"]
    size = module.params["size"]
    opts = shlex.split(module.params["opts"] or "")
    state = module.params["state"]
    force = module.boolean(module.params["force"])
    shrink = module.boolean(module.params["shrink"])
    active = module.boolean(module.params["active"])
    resizefs = module.boolean(module.params["resizefs"])
    thinpool = module.params["thinpool"]
    size_opt = "L"
    size_unit = "m"
    size_operator = None
    snapshot = module.params["snapshot"]
    pvs = module.params["pvs"] or []

    if size:
        # LVEXTEND(8)/LVREDUCE(8) -l, -L options: Check for relative value for resizing
        if size.startswith("+"):
            size_operator = "+"
            size = size[1:]
        elif size.startswith("-"):
            size_operator = "-"
            size = size[1:]
        # LVCREATE(8) does not support [+-]

        # LVCREATE(8)/LVEXTEND(8)/LVREDUCE(8) -l --extents option with percentage
        if "%" in size:
            size_parts = size.split("%", 1)
            size_percent = int(size_parts[0])
            if size_percent > 100:
                module.fail_json(msg="Size percentage cannot be larger than 100%")
            size_whole = size_parts[1]
            if size_whole == "ORIGIN" and snapshot is None:
                module.fail_json(msg="Percentage of ORIGIN supported only for snapshot volumes")
            elif size_whole not in ["VG", "PVS", "FREE", "ORIGIN"]:
                module.fail_json(msg="Specify extents as a percentage of VG|PVS|FREE|ORIGIN")
            size_opt = "l"
            size_unit = ""

        # LVCREATE(8)/LVEXTEND(8)/LVREDUCE(8) -L --size option unit
        if "%" not in size:
            if size[-1].lower() in "bskmgtpe":
                size_unit = size[-1]
                size = size[0:-1]

            try:
                float(size)
                if not size[0].isdigit():
                    raise ValueError()
            except ValueError:
                module.fail_json(msg=f"Bad size specification of '{size}'")

    # when no unit, megabytes by default
    if size_opt == "l":
        unit = "m"
    else:
        unit = size_unit

    vgs = vgs_runner(module)
    lvs = lvs_runner(module)
    lvcreate = lvcreate_runner(module)
    lvremove = lvremove_runner(module)
    lvextend = lvextend_runner(module)
    lvreduce = lvreduce_runner(module)
    lvchange = lvchange_runner(module)

    # Get information on volume group requested
    with vgs("noheadings nosuffix units separator fields vg") as ctx:
        rc, current_vgs, err = ctx.run(units=unit.lower(), separator=";", fields="vg_name,size,free,vg_extent_size")

    if rc != 0:
        if state == "absent":
            module.exit_json(changed=False, stdout=f"Volume group {vg} does not exist.")
        else:
            module.fail_json(msg=f"Volume group {vg} does not exist.", rc=rc, err=err)

    vg_data = parse_vgs(current_vgs)
    this_vg = vg_data[0]

    # Get information on logical volume requested
    with lvs("all noheadings nosuffix units separator fields vg") as ctx:
        rc, current_lvs, err = ctx.run(units=unit.lower(), separator=";", fields="lv_name,size,lv_attr")

    if rc != 0:
        if state == "absent":
            module.exit_json(changed=False, stdout=f"Volume group {vg} does not exist.")
        else:
            module.fail_json(msg=f"Volume group {vg} does not exist.", rc=rc, err=err)

    changed = False

    lv_data = parse_lvs(current_lvs)

    if snapshot:
        # Check snapshot pre-conditions
        for test_lv in lv_data:
            if test_lv["name"] == lv or test_lv["name"] == thinpool:
                if not test_lv["thinpool"] and not thinpool:
                    break
                else:
                    module.fail_json(msg="Snapshots of thin pool LVs are not supported.")
        else:
            module.fail_json(msg=f"Snapshot origin LV {lv} does not exist in volume group {vg}.")
        check_lv = snapshot
    elif thinpool:
        if lv:
            # Check thin volume pre-conditions
            for test_lv in lv_data:
                if test_lv["name"] == thinpool:
                    break
            else:
                module.fail_json(msg=f"Thin pool LV {thinpool} does not exist in volume group {vg}.")
            check_lv = lv
        else:
            check_lv = thinpool
    else:
        check_lv = lv

    for test_lv in lv_data:
        if test_lv["name"] in (check_lv, check_lv.rsplit("/", 1)[-1]):
            this_lv = test_lv
            break
    else:
        this_lv = None

    msg = ""
    if this_lv is None:
        if state == "present":
            if size_operator is not None:
                if size_operator == "-" or (size_whole not in ["VG", "PVS", "FREE", "ORIGIN", None]):
                    module.fail_json(msg=f"Bad size specification of '{size_operator}{size}' for creating LV")
            # Require size argument except for snapshot of thin volumes
            if (lv or thinpool) and not size:
                for test_lv in lv_data:
                    if test_lv["name"] == lv and test_lv["thinvol"] and snapshot:
                        break
                else:
                    module.fail_json(msg="No size given.")

            # create LV
            size_value = f"{size}{size_unit}"
            if snapshot is not None:
                if size_opt == "l":
                    with lvcreate("test yes size_l is_snapshot lv opts vg") as ctx:
                        rc, dummy, err = ctx.run(
                            test=module.check_mode,
                            yes=use_yes,
                            size_l=size_value if size else None,
                            is_snapshot=True,
                            lv=snapshot,
                            opts=opts,
                            vg=[f"{vg}/{lv}"],
                        )
                else:
                    with lvcreate("test yes size_L is_snapshot lv opts vg") as ctx:
                        rc, dummy, err = ctx.run(
                            test=module.check_mode,
                            yes=use_yes,
                            size_L=size_value if size else None,
                            is_snapshot=True,
                            lv=snapshot,
                            opts=opts,
                            vg=[f"{vg}/{lv}"],
                        )
            elif thinpool:
                if lv:
                    if size_opt == "l":
                        module.fail_json(changed=False, msg="Thin volume sizing with percentage not supported.")
                    with lvcreate("test yes lv size_V opts thin vg") as ctx:
                        rc, dummy, err = ctx.run(
                            test=module.check_mode,
                            yes=use_yes,
                            lv=lv,
                            size_V=size_value,
                            opts=opts,
                            vg=[f"{vg}/{thinpool}"],
                        )
                else:
                    if size_opt == "l":
                        with lvcreate("test yes size_l opts thin vg pvs") as ctx:
                            rc, dummy, err = ctx.run(
                                test=module.check_mode,
                                yes=use_yes,
                                size_l=size_value,
                                opts=opts,
                                vg=[f"{vg}/{thinpool}"],
                                pvs=pvs,
                            )
                    else:
                        with lvcreate("test yes size_L opts thin vg pvs") as ctx:
                            rc, dummy, err = ctx.run(
                                test=module.check_mode,
                                yes=use_yes,
                                size_L=size_value,
                                opts=opts,
                                vg=[f"{vg}/{thinpool}"],
                                pvs=pvs,
                            )
            else:
                if size_opt == "l":
                    with lvcreate("test yes lv size_l opts vg pvs") as ctx:
                        rc, dummy, err = ctx.run(
                            test=module.check_mode,
                            yes=use_yes,
                            lv=lv,
                            size_l=size_value,
                            opts=opts,
                            pvs=pvs,
                        )
                else:
                    with lvcreate("test yes lv size_L opts vg pvs") as ctx:
                        rc, dummy, err = ctx.run(
                            test=module.check_mode,
                            yes=use_yes,
                            lv=lv,
                            size_L=size_value,
                            opts=opts,
                            pvs=pvs,
                        )
            if rc == 0:
                changed = True
            else:
                module.fail_json(msg=f"Creating logical volume '{lv}' failed", rc=rc, err=err)
    else:
        if state == "absent":
            # remove LV
            if not force:
                module.fail_json(msg=f"Sorry, no removal of logical volume {this_lv['name']} without force=true.")
            with lvremove("test force lv") as ctx:
                rc, dummy, err = ctx.run(test=module.check_mode, force=True, lv=[f"{vg}/{this_lv['name']}"])
            if rc == 0:
                module.exit_json(changed=True)
            else:
                module.fail_json(msg=f"Failed to remove logical volume {lv}", rc=rc, err=err)

        elif not size:
            pass

        elif size_opt == "l":
            # Resize LV based on % value
            tool = None
            size_free = this_vg["free"]
            if size_whole == "VG" or size_whole == "PVS":
                size_requested = size_percent * this_vg["size"] / 100
            else:  # size_whole == 'FREE':
                size_requested = size_percent * this_vg["free"] / 100

            if size_operator == "+":
                size_requested += this_lv["size"]
            elif size_operator == "-":
                size_requested = this_lv["size"] - size_requested

            # According to latest documentation (LVM2-2.03.11) all tools round down
            size_requested -= size_requested % this_vg["ext_size"]

            if this_lv["size"] < size_requested:
                if (size_free > 0) and (size_free >= (size_requested - this_lv["size"])):
                    tool = "extend"
                else:
                    module.fail_json(
                        msg=(
                            f"Logical Volume {this_lv['name']} could not be extended. Not enough free space left "
                            f"({size_requested - this_lv['size']}{unit} required / {size_free}{unit} available)"
                        )
                    )
            elif shrink and this_lv["size"] > size_requested + this_vg["ext_size"]:  # more than an extent too large
                if size_requested < 1:
                    module.fail_json(msg=f"Sorry, no shrinking of {this_lv['name']} to 0 permitted.")
                elif not force:
                    module.fail_json(msg=f"Sorry, no shrinking of {this_lv['name']} without force=true")
                else:
                    tool = "reduce"

            if tool:
                size_value = f"{size_operator or ''}{size}{size_unit}"
                lv_name = f"{vg}/{this_lv['name']}"
                rc, out, err = _run_resize(
                    module, tool == "extend", lvextend, lvreduce, size_opt, size_value, resizefs, lv_name, pvs
                )
                if "Reached maximum COW size" in out:
                    module.fail_json(msg=f"Unable to resize {lv} to {size}{size_unit}", rc=rc, err=err, out=out)
                elif rc == 0:
                    changed = True
                    msg = f"Volume {this_lv['name']} resized to {size_requested}{unit}"
                elif "matches existing size" in err or "matches existing size" in out:
                    module.exit_json(changed=False, vg=vg, lv=this_lv["name"], size=this_lv["size"])
                elif "not larger than existing size" in err or "not larger than existing size" in out:
                    module.exit_json(
                        changed=False,
                        vg=vg,
                        lv=this_lv["name"],
                        size=this_lv["size"],
                        msg="Original size is larger than requested size",
                        err=err,
                    )
                else:
                    module.fail_json(msg=f"Unable to resize {lv} to {size}{size_unit}", rc=rc, err=err)

        else:
            # resize LV based on absolute values
            tool = None
            if float(size) > this_lv["size"] or size_operator == "+":
                tool = "extend"
            elif shrink and float(size) < this_lv["size"] or size_operator == "-":
                if float(size) == 0:
                    module.fail_json(msg=f"Sorry, no shrinking of {this_lv['name']} to 0 permitted.")
                if not force:
                    module.fail_json(msg=f"Sorry, no shrinking of {this_lv['name']} without force=true.")
                else:
                    tool = "reduce"

            if tool:
                size_value = f"{size_operator or ''}{size}{size_unit}"
                lv_name = f"{vg}/{this_lv['name']}"
                rc, out, err = _run_resize(
                    module, tool == "extend", lvextend, lvreduce, size_opt, size_value, resizefs, lv_name, pvs
                )
                if "Reached maximum COW size" in out:
                    module.fail_json(msg=f"Unable to resize {lv} to {size}{size_unit}", rc=rc, err=err, out=out)
                elif rc == 0:
                    changed = True
                elif "matches existing size" in err or "matches existing size" in out:
                    module.exit_json(changed=False, vg=vg, lv=this_lv["name"], size=this_lv["size"])
                elif "not larger than existing size" in err or "not larger than existing size" in out:
                    module.exit_json(
                        changed=False,
                        vg=vg,
                        lv=this_lv["name"],
                        size=this_lv["size"],
                        msg="Original size is larger than requested size",
                        err=err,
                    )
                else:
                    module.fail_json(msg=f"Unable to resize {lv} to {size}{size_unit}", rc=rc, err=err)

    if this_lv is not None:
        if active:
            with lvchange("active lv") as ctx:
                rc, dummy, err = ctx.run(active=True, lv=[f"{vg}/{this_lv['name']}"])
            if rc == 0:
                module.exit_json(
                    changed=(not this_lv["active"]) or changed, vg=vg, lv=this_lv["name"], size=this_lv["size"]
                )
            else:
                module.fail_json(msg=f"Failed to activate logical volume {lv}", rc=rc, err=err)
        else:
            with lvchange("active lv") as ctx:
                rc, dummy, err = ctx.run(active=False, lv=[f"{vg}/{this_lv['name']}"])
            if rc == 0:
                module.exit_json(changed=this_lv["active"] or changed, vg=vg, lv=this_lv["name"], size=this_lv["size"])
            else:
                module.fail_json(msg=f"Failed to deactivate logical volume {lv}", rc=rc, err=err)

    module.exit_json(changed=changed, msg=msg)


if __name__ == "__main__":
    main()
